/* This file has been autogenerated with make -C d themes.c */
static const char * const theme_ayu = \
	"# verified by pancake @ 2021-12-21\n"
	"ecd\n"
	"ec ai.exec rgb:6d6\n"
	"ec ai.read rgb:66d\n"
	"ec ai.write rgb:d66\n"
	"ec args rgb:2bc\n"
	"ec b0x00 rgb:789\n"
	"ec b0x7f rgb:677\n"
	"ec b0xff rgb:bbb             # ff bytes\n"
	"ec btext rgb:5de             # text section opcode? first byte of opcode\n"
	"ec comment rgb:cd4 #95e6cb           # r2 renerated comments\n"
	"ec creg rgb:ff7733              # modified registries\n"
	"ec flag rgb:5de              # flags. usually apprears above the current seek\n"
	"ec fline rgb:55b4d4             # function line. the boundary of a function\n"
	"ec flow rgb:cd4              # lines that show jump destinations\n"
	"ec flow2 rgb:ddd\n"
	"ec fname rgb:5de # function name\n"
	"ec graph.traced rgb:2eb\n"
	"ec graph.box rgb:5c6773         # unselected boxes in visual graph mode\n"
	"ec graph.box2 rgb:ff6a00        # currently selected box in visual graph mode\n"
	"ec graph.box3 rgb:66d9ef\n"
	"ec graph.box4 rgb:66d9ef\n"
	"ec graph.current rgb:fff\n"
	"ec graph.false rgb:f07178       # false line on visual graph mode\n"
	"ec graph.true rgb:bae67e        # true line on visual graph mode\n"
	"ec graph.trufae rgb:55b4d4      # lines in graph that does not have true or false\n"
	"ec help rgb:c2d94c              # color of help explanation text\n"
	"ec input rgb:888\n"
	"ec jmp rgb:c2d94c               # jmp instructions\n"
	"ec label rgb:5de\n"
	"ec linehl rgb:42505e            # highlight color when using asm.highlight\n"
	"ec mov white               # mov instructions\n"
	"ec nop rgb:a7c               # nop instructions\n"
	"ec num rgb:f85               # numbers\n"
	"ec offset rgb:d74            # offset in visual mode\n"
	"ec other rgb:888\n"
	"ec prompt rgb:5de               # color for main r2 prompt\n"
	"ec call rgb:f73              # push instructions\n"
	"ec reg yellow\n"
	"ec ret rgb:d66                  # ret opcode\n"
	"ec swi rgb:d66\n"
	"ec trap rgb:d66\n"
	"ec cmp rgb:5de\n"
	"ec math rgb:789\n"
	"ec bin rgb:789\n"
	"ec push rgb:5de\n"
	"ec pop rgb:5de\n"
	"ec cjmp rgb:c2d94c\n"
	"ec ucall rgb:5ccfe6\n"
	"ec ujmp rgb:c2d94c\n"
	"ec usrcmt rgb:fa6e32\n"
	"\n"
	"ec var rgb:95e6cb\n"
	"ec var.type rgb:ffd580\n"
	"ec var.addr rgb:f28779\n"
	"ec var.name rgb:95e6cb\n"
	"\n"
	"ec widget.bg rgb:888 rgb:000\n"
	"ec widget.sel rgb:5de rgb:333\n"
;
static const char * const theme_basic = \
	"# optimized for windows\n"
	"ecd\n"
	"ec fname cyan\n"
	"ec label cyan\n"
	"ec fline cyan\n"
	"ec flag green\n"
	"ec offset white\n"
	"ec comment red\n"
	"ec usrcmt green\n"
	"ec flow green\n"
	"ec flow2 green\n"
	"ec b0x00 white\n"
	"ec b0xff red\n"
	"ec bin gray\n"
	"ec b0x7f white\n"
	"ec other gray\n"
	"ec btext cyan\n"
	"ec trap red\n"
	"ec ucall green\n"
	"ec ujmp green\n"
	"ec args green\n"
	"ec help yellow\n"
	"ec prompt cyan\n"
	"ec mov white\n"
	"ec math gray\n"
	"ec push cyan\n"
	"ec reg yellow\n"
	"ec creg yellow\n"
	"ec num cyan\n"
	"ec invalid red\n"
	"ec swi cyan\n"
	"ec cmp white\n"
	"ec call green\n"
	"ec cjmp green\n"
	"ec jmp green\n"
	"ec pop cyan\n"
	"ec linehl rgb:004\n"
	"ec graph.true green\n"
	"ec graph.false red\n"
	"ec graph.trufae blue\n"
	"ec graph.current blue\n"
	"ec graph.traced blue\n"
	"ec graph.box gray\n"
	"ec graph.box2 blue\n"
	"ec graph.box3 gray\n"
	"ec graph.box4 gray\n"
	"\n"
	"ec var white\n"
	"ec var.type blue\n"
	"ec var.addr cyan\n"
	"ec var.name white\n"
	"\n"
	"ec widget.bg white gray\n"
	"ec widget.sel black cyan\n"
;
static const char * const theme_behelit = \
	"ecd\n"
	"ec b0x7f rgb:ff5f5f\n"
	"ec b0xff rgb:ff005f\n"
	"ec b0x00 rgb:262626\n"
	"ec btext rgb:af87ff\n"
	"ec other rgb:4e4e4e\n"
	"ec offset rgb:4e4e4e\n"
	"ec flow rgb:5f87ff\n"
	"ec flow2 rgb:5f87ff\n"
	"ec input rgb:5fff5f\n"
	"ec prompt rgb:5f87ff\n"
	"ec args rgb:ff005f\n"
	"ec help rgb:af87d7\n"
	"ec comment rgb:ff005f\n"
	"ec usrcmt rgb:000000 rgb:5f87ff\n"
	"ec flag rgb:5f87ff\n"
	"ec fline rgb:af87ff\n"
	"ec fname rgb:5f87ff\n"
	"ec floc rgb:5f87ff\n"
	"ec label rgb:5f87ff\n"
	"ec mov rgb:5f87ff\n"
	"ec call rgb:000000 rgb:00ff5f\n"
	"ec jmp rgb:000000 rgb:5f5fff\n"
	"ec cjmp rgb:000000 rgb:5f87ff\n"
	"ec cmp rgb:ff005f\n"
	"ec nop rgb:4e4e4e\n"
	"ec math rgb:ffff5f\n"
	"ec bin rgb:ff5f5f\n"
	"ec ret rgb:ff005f\n"
	"ec push rgb:af87ff\n"
	"ec pop rgb:af87ff\n"
	"ec swi rgb:8787ff\n"
	"ec crypto rgb:000000 rgb:ff5f5f\n"
	"ec trap rgb:000000 rgb:ff005f\n"
	"ec ucall rgb:000000 rgb:00ff5f\n"
	"ec ujmp rgb:000000 rgb:5f87ff\n"
	"\n"
	"ec invalid rgb:000000 rgb:ff005f\n"
	"ec num rgb:af87ff rgb:121212\n"
	"ec reg rgb:5fff5f rgb:121212\n"
	"ec creg rgb:5f87ff\n"
	"ec ai.read rgb:00ff5f\n"
	"ec ai.write rgb:5f87ff\n"
	"ec ai.exec rgb:ff005f\n"
	"ec ai.seq rgb:af87ff\n"
	"ec ai.ascii rgb:ff5f5f\n"
	"ec graph.box rgb:00ff5f\n"
	"ec graph.box2 rgb:5f87ff\n"
	"ec graph.box3 rgb:af87ff\n"
	"ec graph.box4 rgb:ff005f\n"
	"ec graph.true rgb:00ff5f\n"
	"ec graph.false rgb:ff005f\n"
	"ec graph.trufae rgb:5f87ff\n"
	"ec graph.current rgb:5f87ff\n"
	"ec graph.traced rgb:ffff5f\n"
	"ec linehl rgb:1f1f1f\n"
	"\n"
	"# background of the terminal must be #121212\n"
	"# foreground of the terminal must be #5fff5f\n"
	"# xterm -bg rgb:12/12/12 -fg rgb:5f/ff/5f\n"
	"\n"
	"ec var rgb:ff005f\n"
	"ec var.type blue\n"
	"ec var.addr rgb:5f87ff\n"
	"ec var.name rgb:ff005f\n"
	"\n"
	"ec widget.bg rgb:68f rgb:555\n"
	"ec widget.sel black rgb:0f6\n"
;
static const char * const theme_blood = \
	"ecd\n"
	"ec b0x00 rgb:d00\n"
	"ec b0x7f rgb:900\n"
	"ec b0xff rgb:500\n"
	"ec args rgb:f00\n"
	"ec bin rgb:f00\n"
	"ec btext rgb:f00\n"
	"ec call rgb:faa\n"
	"ec cjmp rgb:faa\n"
	"ec ujmp rgb:faa\n"
	"ec jmp rgb:faa\n"
	"ec ret rgb:faa\n"
	"ec cmp rgb:a00\n"
	"ec comment rgb:700\n"
	"ec usrcmt rgb:f00\n"
	"ec reg rgb:f00\n"
	"ec creg rgb:f00\n"
	"ec flag rgb:f00\n"
	"ec fline rgb:a00\n"
	"ec flow rgb:f00\n"
	"ec flow2 rgb:f00\n"
	"ec fname rgb:f00\n"
	"ec help rgb:f00\n"
	"ec input rgb:f00\n"
	"ec label rgb:f00\n"
	"ec math rgb:700\n"
	"ec mov rgb:d00\n"
	"ec nop rgb:300\n"
	"ec num rgb:f00\n"
	"ec offset rgb:500\n"
	"ec other rgb:900\n"
	"ec pop rgb:900\n"
	"ec prompt rgb:f00\n"
	"ec push rgb:d00\n"
	"ec swi rgb:f00\n"
	"ec trap rgb:f00\n"
	"ec ucall rgb:f00 rgb:300\n"
	"ec graph.true rgb:f00\n"
	"ec graph.false rgb:700\n"
	"ec graph.trufae rgb:d00\n"
	"ec graph.current rgb:f00\n"
	"ec graph.traced rgb:f00\n"
	"ec graph.box rgb:f00\n"
	"ec graph.box2 rgb:f00\n"
	"ec graph.box3 rgb:000\n"
	"ec graph.box4 rgb:000\n"
	"ec gui.cflow rgb:f00\n"
	"ec gui.dataoffset rgb:f00\n"
	"ec crypto rgb:f00\n"
	"ec floc rgb:700\n"
	"ec linehl rgb:111\n"
	"\n"
	"ec var rgb:f00\n"
	"ec var.type rgb:700\n"
	"ec var.addr rgb:500\n"
	"ec var.name rgb:f00\n"
	"\n"
	"ec ai.read rgb:eee\n"
	"ec ai.write rgb:d00\n"
	"ec ai.exec rgb:f00\n"
	"ec ai.seq rgb:f00\n"
	"ec ai.ascii rgb:f00\n"
	"\n"
	"ec widget.bg rgb:f00 rgb:300\n"
	"ec widget.sel rgb:300 rgb:f00\n"
;
static const char * const theme_bluy = \
	"ecd\n"
	"ec comment rgb:fc1\n"
	"ec fname rgb:fc1\n"
	"ec fline rgb:f91\n"
	"ec flag rgb:fc1\n"
	"ec label rgb:fc1\n"
	"# help messages\n"
	"ec help rgb:3cf\n"
	"ec args rgb:09f\n"
	"ec flow rgb:e71\n"
	"ec prompt rgb:09f\n"
	"ec offset rgb:666\n"
	"ec input rgb:fc1\n"
	"ec invalid rgb:e34\n"
	"ec other rgb:09f\n"
	"ec b0x00 rgb:99a\n"
	"ec b0x7f rgb:09f\n"
	"ec b0xff rgb:555\n"
	"ec btext rgb:3cf\n"
	"ec math rgb:09f\n"
	"ec bin rgb:99a\n"
	"ec push rgb:09f\n"
	"ec pop rgb:09f\n"
	"ec crypto rgb:fff\n"
	"ec jmp rgb:e71\n"
	"ec cjmp rgb:e71\n"
	"ec call magenta\n"
	"ec nop rgb:7c5\n"
	"ec ret rgb:e34\n"
	"ec trap rgb:e34\n"
	"ec swi rgb:e34\n"
	"ec cmp rgb:09f\n"
	"ec reg rgb:3cf\n"
	"ec creg rgb:09f\n"
	"ec num rgb:fc1\n"
	"ec mov rgb:99a\n"
	"ec ai.read rgb:e33\n"
	"ec ai.write rgb:33e\n"
	"ec ai.exec rgb:3cf\n"
	"ec ai.seq rgb:09f\n"
	"ec ai.ascii rgb:fff\n"
	"ec gui.cflow rgb:09f\n"
	"ec gui.dataoffset rgb:99a\n"
	"ec gui.background rgb:7c5\n"
	"ec gui.background2 rgb:e34\n"
	"ec gui.border rgb:23f\n"
	"ec graph.current rgb:aaf\n"
	"\n"
	"ec var rgb:555\n"
	"ec var.name rgb:09f\n"
	"ec var.type rgb:99a\n"
	"ec var.addr rgb:666\n"
	"\n"
	"# ec graph.box4=white\n"
	"# \"\"e graph.gv.node=fillcolor=\"#303030\" color=lightgray bgcolor=red fontcolor=white penwidth=2 style=filled fontname=\"Courier Bold\" fontsize=14 shape=box\n"
	"# e graph.gv.graph=bgcolor=black\n"
;
static const char * const theme_bobcrawl = \
	"ecd\n"
	"ec call rgb:ffff00\n"
	"ec ucall rgb:ffff00\n"
	"ec comment rgb:32302f\n"
	"ec usrcmt rgb:ffc5f3\n"
	"ec ret rgb:fb4934\n"
	"ec args rgb:a89984\n"
	"ec fname rgb:fad200\n"
	"ec fline rgb:458588\n"
	"ec floc rgb:00ffff\n"
	"ec flag rgb:d75f00\n"
	"ec label rgb:8787d7\n"
	"ec help rgb:ff5f87\n"
	"ec flow2 rgb:0000ff\n"
	"ec prompt rgb:d75f00\n"
	"ec offset rgb:afafaf\n"
	"ec input rgb:d75f00\n"
	"ec invalid rgb:001f00\n"
	"ec other rgb:5f8787\n"
	"ec b0x00 rgb:005f5f\n"
	"ec b0x7f rgb:005f5f\n"
	"ec b0xff rgb:5f8787\n"
	"ec math rgb:00afd7\n"
	"ec bin rgb:87afaf\n"
	"ec btext rgb:afafaf\n"
	"ec push rgb:ff5f87\n"
	"ec pop rgb:87afaf\n"
	"ec crypto rgb:0000ff\n"
	"ec jmp rgb:ff5f87\n"
	"ec cjmp rgb:afaf00\n"
	"ec ujmp rgb:ff5f87\n"
	"ec nop rgb:5f8787\n"
	"ec trap rgb:ff0000\n"
	"ec swi rgb:ff0000\n"
	"ec cmp rgb:d75f00\n"
	"ec reg rgb:73adad\n"
	"ec creg rgb:73adad\n"
	"ec num rgb:d25032\n"
	"ec mov rgb:ebebeb\n"
	"ec linehl rgb:00003f\n"
	"ec ai.read rgb:00ff00\n"
	"ec ai.write rgb:0000ff\n"
	"ec ai.exec rgb:ff0000\n"
	"ec ai.seq rgb:ff00ff\n"
	"ec ai.ascii rgb:ffff00\n"
	"ec graph.box=rgb:9e9e9e\n"
	"ec graph.box2 rgb:5f87d7\n"
	"ec graph.box3 rgb:af5f87\n"
	"ec graph.box4 rgb:303030\n"
	"ec graph.trufae rgb:005fff\n"
	"ec graph.current rgb:c68e00\n"
	"ec graph.true rgb:009100\n"
	"ec graph.false rgb:bc0000\n"
	"ec graph.traced rgb:ffff00\n"
	"ec gui.cflow rgb:ffff00\n"
	"ec gui.dataoffset rgb:ffff00\n"
	"ec gui.background rgb:303030\n"
	"ec gui.background2 rgb:ffffff\n"
	"ec gui.border rgb:5fd700\n"
	"\n"
	"ec var rgb:64604f\n"
	"ec var.type rgb:00afd7\n"
	"ec var.addr rgb:afafaf\n"
	"ec var.name rgb:64604f\n"
	"\n"
	"ec ai.read rgb:a55f00\n"
	"ec ai.write rgb:d75f00\n"
	"ec ai.exec rgb:ff5f87\n"
	"ec ai.seq rgb:9dd600\n"
	"ec ai.ascii rgb:d25032\n"
	"\n"
	"ec widget.bg rgb:099 rgb:444\n"
	"ec widget.sel rgb:000 yellow\n"
;
static const char * const theme_bold = \
	"ecd\n"
	"ec b0x00 rgb:ccc\n"
	"ec b0x7f rgb:ccc\n"
	"ec b0xff red\n"
	"ec btext rgb:fff . bold\n"
	"\n"
	"ec args white\n"
	"ec bin white\n"
	"ec cmp white\n"
	"ec usrcmt black white bold\n"
	"ec reg green . bold\n"
	"ec creg rgb:f72\n"
	"ec flag yellow . bold\n"
	"ec fline white\n"
	"\n"
	"ec flow yellow\n"
	"ec flow2 white\n"
	"\n"
	"ec fname black green bold\n"
	"ec help yellow . bold\n"
	"ec input yellow . bold\n"
	"ec label white . bold\n"
	"ec math white\n"
	"ec mov white\n"
	"ec nop rgb:ccc\n"
	"\n"
	"ec comment yellow . bold\n"
	"ec call black green bold\n"
	"ec ucall black green bold\n"
	"ec num rgb:fff . bold\n"
	"ec jmp yellow . bold\n"
	"ec cjmp yellow . bold\n"
	"ec ujmp yellow . bold\n"
	"\n"
	"ec offset white . bold\n"
	"ec other rgb:999\n"
	"ec pop rgb:999\n"
	"\n"
	"ec prompt black yellow bold\n"
	"\n"
	"ec graph.box white\n"
	"ec graph.box2 green\n"
	"ec graph.box3 rgb:000\n"
	"ec graph.box4 rgb:000\n"
	"\n"
	"ec graph.true green . bold\n"
	"ec graph.false red\n"
	"\n"
	"ec push white\n"
	"ec ret rgb:777\n"
	"ec swi black red bold\n"
	"ec trap red\n"
	"ec graph.trufae rgb:ddd\n"
	"ec graph.current rgb:f72\n"
	"ec graph.traced rgb:f00\n"
	"ec gui.cflow rgb:f72\n"
	"ec gui.dataoffset rgb:f72\n"
	"ec crypto rgb:f72\n"
	"ec floc rgb:777\n"
	"ec invalid red\n"
	"ec linehl rgb:008\n"
	"\n"
	"ec var yellow . bold\n"
	"ec var.type blue . bold\n"
	"ec var.addr cyan . bold\n"
	"ec var.name yellow\n"
	"\n"
	"ec ai.read white . bold\n"
	"ec ai.write white . bold\n"
	"ec ai.exec red . bold\n"
	"ec ai.seq green . bold\n"
	"ec ai.ascii yellow . bold\n"
	"\n"
	"ec widget.bg yellow black\n"
	"ec widget.sel black yellow\n"
;
static const char * const theme_bright = \
	"# r2 default palette but with bright colors for all\n"
	"ecd\n"
	"ec b0x00 rgb:00ff00\n"
	"ec b0x7f rgb:00ffff\n"
	"ec b0xff rgb:ff0000\n"
	"ec args rgb:ffff00\n"
	"ec bin rgb:00ffff\n"
	"ec btext rgb:ffff00\n"
	"ec call rgb:00ff00 . bold\n"
	"ec cjmp rgb:00ff00\n"
	"ec cmp rgb:00ffff\n"
	"ec comment rgb:ff0000\n"
	"ec usrcmt rgb:ffffff\n"
	"ec creg rgb:00ffff\n"
	"ec flag rgb:00ffff\n"
	"ec fline rgb:00ffff\n"
	"ec floc rgb:00ffff\n"
	"ec flow rgb:00ffff\n"
	"ec flow2 rgb:0000ff\n"
	"ec fname rgb:ff0000\n"
	"ec help rgb:00ff00\n"
	"ec input rgb:ffffff\n"
	"ec invalid rgb:ff0000 . bold\n"
	"ec jmp rgb:00ff00\n"
	"ec label rgb:00ffff\n"
	"ec math rgb:ffff00\n"
	"ec mov rgb:ffffff\n"
	"ec nop rgb:0000ff\n"
	"ec num rgb:ffff00\n"
	"ec offset rgb:00ff00\n"
	"ec other rgb:ffffff\n"
	"ec pop rgb:ff00ff . bold\n"
	"ec prompt rgb:ffff00\n"
	"ec push rgb:ff00ff\n"
	"ec crypto rgb:ffff00\n"
	"ec reg rgb:00ffff\n"
	"ec ret rgb:ff0000\n"
	"ec swi rgb:ff00ff\n"
	"ec trap rgb:ff0000 . bold\n"
	"ec ucall rgb:00ff00\n"
	"ec ujmp rgb:00ff00\n"
	"\n"
	"ec ai.read rgb:00ff00\n"
	"ec ai.write rgb:0000ff\n"
	"ec ai.exec rgb:ff0000\n"
	"ec ai.seq rgb:ff00ff\n"
	"ec ai.ascii rgb:ffff00\n"
	"\n"
	"ec gui.cflow rgb:ffff00\n"
	"ec gui.dataoffset rgb:ffff00\n"
	"ec gui.background rgb:000000\n"
	"ec gui.background2 rgb:ffffff\n"
	"ec gui.border rgb:000000\n"
	"ec wordhl rgb:ff0000\n"
	"ec linehl rgb:000080\n"
	"\n"
	"ec var rgb:ffffff\n"
	"ec var.type rgb:ffff00\n"
	"ec var.addr rgb:00ffff\n"
	"ec var.name rgb:ffffff\n"
	"\n"
	"ec widget.bg rgb:303030\n"
	"ec widget.sel rgb:ff0000\n"
	"\n"
	"ec graph.box rgb:0000ff\n"
	"ec graph.box2 rgb:00ff00\n"
	"ec graph.box3 rgb:ff00ff\n"
	"ec graph.box4 rgb:808080\n"
	"ec graph.true rgb:00ff00\n"
	"ec graph.false rgb:ff0000\n"
	"ec graph.trufae rgb:0000ff    # single jump\n"
	"ec graph.traced rgb:ffff00\n"
	"ec graph.current rgb:0000ff\n"
;
static const char * const theme_cga = \
	"ecd\n"
	"ec comment rgb:0ff\n"
	"ec usrcmt rgb:f0f\n"
	"ec mov white\n"
	"ec fname rgb:f00\n"
	"ec args white\n"
	"ec floc rgb:f0f\n"
	"ec fline rgb:f0f\n"
	"ec flag rgb:f0f\n"
	"ec invalid rgb:f0f\n"
	"ec trap rgb:f0f\n"
	"ec swi rgb:f0f\n"
	"ec ret rgb:f0f\n"
	"ec input rgb:0ff\n"
	"ec fname rgb:f0f\n"
	"ec label rgb:f0f\n"
	"ec flow rgb:f0f\n"
	"ec flow2 rgb:f0f\n"
	"ec prompt rgb:0ff\n"
	"ec offset rgb:0ff\n"
	"ec other rgb:fff\n"
	"ec b0x00 rgb:fff\n"
	"ec b0x7f rgb:f0f\n"
	"ec b0xff rgb:0ff\n"
	"ec math rgb:f0f\n"
	"ec crypto rgb:f0f\n"
	"ec bin rgb:f0f\n"
	"ec btext rgb:f0f\n"
	"ec push rgb:0ff\n"
	"ec pop rgb:0ff\n"
	"ec jmp rgb:f0f\n"
	"ec cjmp rgb:f0f\n"
	"ec call rgb:f0f\n"
	"ec nop rgb:0ff\n"
	"ec trap rgb:f0f\n"
	"ec cmp rgb:0ff\n"
	"ec reg rgb:f0f\n"
	"ec creg rgb:f0f\n"
	"ec num rgb:0ff\n"
	"ec help rgb:0ff\n"
	"ec trap rgb:000 rgb:f0f\n"
	"ec ucall rgb:f0f\n"
	"ec ujmp rgb:f0f\n"
	"ec linehl rgb:008\n"
	"\n"
	"ec graph.box rgb:f0f\n"
	"ec graph.box2 rgb:f0f\n"
	"ec graph.box3 rgb:f0f\n"
	"ec graph.box4 rgb:f0f\n"
	"ec graph.true rgb:f0f\n"
	"ec graph.false rgb:0ff\n"
	"ec graph.trufae rgb:f0f\n"
	"ec graph.current rgb:0ff\n"
	"ec graph.traced rgb:f0f\n"
	"\n"
	"ec gui.cflow rgb:0ff\n"
	"ec gui.dataoffset rgb:0ff\n"
	"ec gui.background rgb:0ff\n"
	"ec gui.background2 rgb:0ff\n"
	"ec gui.border rgb:0ff\n"
	"\n"
	"ec ai.read rgb:0ff\n"
	"ec ai.write rgb:f0f\n"
	"ec ai.exec rgb:f0f\n"
	"ec ai.seq rgb:f0f\n"
	"ec ai.ascii rgb:f0f\n"
	"\n"
	"ec var rgb:f0f\n"
	"ec var.type rgb:f0f\n"
	"ec var.addr rgb:0ff\n"
	"ec var.name rgb:f0f\n"
	"\n"
	"ec widget.bg rgb:f0f rgb:000\n"
	"ec widget.sel rgb:0ff rgb:000\n"
;
static const char * const theme_chrome = \
	"ecd\n"
	"ec b0x00 rgb:ccc\n"
	"ec b0x7f rgb:777\n"
	"ec b0xff rgb:333\n"
	"ec args rgb:fff\n"
	"ec bin rgb:fff\n"
	"ec btext rgb:fff\n"
	"ec call white\n"
	"ec cjmp white\n"
	"ec cmp rgb:aaa\n"
	"ec comment white\n"
	"ec usrcmt rgb:333 white\n"
	"ec reg rgb:fff\n"
	"ec creg rgb:fff\n"
	"ec flag rgb:fff\n"
	"ec fline rgb:aaa\n"
	"ec flow rgb:fff\n"
	"ec flow2 rgb:fff\n"
	"ec fname rgb:fff\n"
	"ec help rgb:fff\n"
	"ec input rgb:fff\n"
	"ec jmp rgb:fff\n"
	"ec label rgb:fff\n"
	"ec math rgb:777\n"
	"ec mov rgb:ddd\n"
	"ec nop rgb:333\n"
	"ec num rgb:fff\n"
	"ec offset rgb:555\n"
	"ec other rgb:999\n"
	"ec pop rgb:999\n"
	"ec prompt rgb:fff\n"
	"ec push rgb:ddd\n"
	"ec ret rgb:777\n"
	"ec swi rgb:fff\n"
	"ec trap rgb:fff\n"
	"ec ucall rgb:fff rgb:333\n"
	"ec ujmp rgb:fff\n"
	"ec graph.true rgb:fff\n"
	"ec graph.false rgb:777\n"
	"ec graph.trufae rgb:ddd\n"
	"ec graph.current rgb:fff\n"
	"ec graph.traced rgb:f00\n"
	"ec graph.box rgb:fff\n"
	"ec graph.box2 rgb:fff\n"
	"ec graph.box3 rgb:000\n"
	"ec graph.box4 rgb:000\n"
	"ec gui.cflow rgb:fff\n"
	"ec gui.dataoffset rgb:fff\n"
	"ec crypto rgb:fff\n"
	"ec floc rgb:777\n"
	"ec linehl rgb:111\n"
	"\n"
	"ec var rgb:fff\n"
	"ec var.type rgb:777\n"
	"ec var.addr rgb:555\n"
	"ec var.name rgb:fff\n"
	"\n"
	"ec ai.read rgb:eee\n"
	"ec ai.write rgb:ddd\n"
	"ec ai.exec rgb:fff\n"
	"ec ai.seq rgb:fff\n"
	"ec ai.ascii rgb:fff\n"
	"\n"
	"ec widget.bg rgb:fff rgb:333\n"
	"ec widget.sel rgb:333 rgb:fff\n"
;
static const char * const theme_consonance = \
	"# background of the terminal must be 134 o 033\n"
	"# xterm -fn 10x20 -bg rgb:20/20/20 -fg rgb:d0/d0/d0\n"
	"# ref: https://github.com/eugeneching/ida-consonance/blob/master/ida-consonance.clr\n"
	"ecd\n"
	"ec b0x00 rgb:9aa\n"
	"ec b0x7f rgb:577\n"
	"ec b0xff rgb:8cf\n"
	"ec bin rgb:fff\n"
	"ec btext rgb:fff\n"
	"ec ucall rgb:2fa\n"
	"ec call rgb:f2a\n"
	"ec ujmp rgb:4cf\n"
	"ec cjmp rgb:4cf\n"
	"ec jmp rgb:4cf\n"
	"ec cmp rgb:fff\n"
	"ec comment rgb:fcc\n"
	"ec usrcmt rgb:f2a\n"
	"ec ret rgb:4cf\n"
	"ec fline rgb:fff\n"
	"ec flow rgb:789\n"
	"ec flow2 rgb:888\n"
	"ec fname rgb:ee0\n"
	"ec label rgb:ee0\n"
	"ec math rgb:fff\n"
	"ec mov rgb:fff\n"
	"ec nop rgb:888\n"
	"ec num rgb:fff\n"
	"ec offset rgb:888\n"
	"ec other rgb:577\n"
	"ec pop rgb:fff\n"
	"ec prompt rgb:4cf\n"
	"ec push rgb:fff\n"
	"ec reg rgb:789\n"
	"ec args rgb:789\n"
	"ec help rgb:ee0\n"
	"ec input rgb:c41\n"
	"ec trap rgb:d41\n"
	"ec swi rgb:f2a\n"
	"ec creg rgb:789\n"
	"ec flag rgb:4cf\n"
	"ec linehl rgb:004\n"
	"\n"
	"ec graph.box rgb:789\n"
	"ec graph.box2 rgb:d41\n"
	"ec graph.box3 rgb:af2\n"
	"ec graph.box4 rgb:af2\n"
	"ec graph.true rgb:4f2\n"
	"ec graph.false rgb:d41\n"
	"ec graph.trufae rgb:4cf\n"
	"ec graph.current rgb:af2\n"
	"ec graph.traced rgb:090\n"
	"\n"
	"ec var rgb:fcc\n"
	"ec var.type rgb:f2a\n"
	"ec var.addr white\n"
	"ec var.name rgb:fcc\n"
	"\n"
	"ec ai.read rgb:fcc\n"
	"ec ai.write rgb:f2a\n"
	"ec ai.exec rgb:f2a\n"
	"ec ai.seq rgb:f0f\n"
	"ec ai.ascii rgb:fff\n"
	"\n"
	"ec widget.bg rgb:4cf rgb:344\n"
	"ec widget.sel black rgb:4cf\n"
;
static const char * const theme_dark = \
	"ecd\n"
	"ec b0x00 rgb:444\n"
	"ec help rgb:366\n"
	"ec comment rgb:99a\n"
	"ec usrcmt rgb:368\n"
	"ec input rgb:bbb\n"
	"ec b0x7f rgb:555\n"
	"ec b0xff rgb:666\n"
	"ec bin rgb:f90\n"
	"ec btext rgb:366\n"
	"ec call rgb:44f rgb:123\n"
	"ec cjmp rgb:368\n"
	"ec cmp rgb:666\n"
	"ec fline rgb:366\n"
	"ec flag rgb:99a\n"
	"ec flow rgb:bbb\n"
	"ec flow2 rgb:99a\n"
	"ec jmp rgb:368\n"
	"ec fname rgb:99a\n"
	"ec label rgb:99a\n"
	"ec math rgb:366\n"
	"ec bin rgb:366\n"
	"ec crypto rgb:366\n"
	"ec mov rgb:666\n"
	"ec nop rgb:444\n"
	"ec num rgb:99a\n"
	"ec offset rgb:366\n"
	"ec other rgb:bbb\n"
	"ec prompt rgb:366\n"
	"ec push rgb:99a\n"
	"ec pop rgb:99a\n"
	"ec reg rgb:368\n"
	"ec creg rgb:368\n"
	"ec trap rgb:000 rgb:123\n"
	"ec swi rgb:99a\n"
	"ec ucall rgb:44f rgb:123\n"
	"ec ujmp rgb:368\n"
	"ec invalid rgb:rgb:44f rgb:123\n"
	"ec linehl rgb:004\n"
	"ec graph.box rgb:040\n"
	"ec graph.box2 rgb:040\n"
	"ec graph.box3 rgb:040\n"
	"ec graph.box4 rgb:040\n"
	"ec graph.true rgb:06c\n"
	"ec graph.false rgb:035\n"
	"ec graph.trufae rgb:06c\n"
	"ec graph.current rgb:99a\n"
	"ec graph.traced rgb:bbb\n"
	"\n"
	"ec var rgb:99a\n"
	"ec var.type rgb:636\n"
	"ec var.addr rgb:366\n"
	"ec var.name rgb:99a\n"
	"\n"
	"ec ai.read rgb:266\n"
	"ec ai.write rgb:366\n"
	"ec ai.exec rgb:368\n"
	"ec ai.seq rgb:d38\n"
	"ec ai.ascii rgb:99a\n"
	"\n"
	"ec widget.bg rgb:366 black\n"
	"ec widget.sel rgb:44f rgb:123\n"
;
static const char * const theme_darkda = \
	"ecd\n"
	"ec b0x00 rgb:7c7c61\n"
	"ec b0x7f rgb:aaff00\n"
	"ec b0xff rgb:aaff00\n"
	"ec other rgb:ff7f00\n"
	"ec bin rgb:ff5afd\n"
	"ec btext rgb:617c7c\n"
	"ec call rgb:ffd200\n"
	"ec cjmp rgb:9dd600\n"
	"ec jmp rgb:9dd600\n"
	"ec cmp rgb:00ffff\n"
	"ec comment rgb:82607e\n"
	"ec usrcmt rgb:f3c5ff\n"
	"ec ret rgb:ffaa7f\n"
	"ec num rgb:d25032\n"
	"ec fline rgb:2bc\n"
	"ec flow rgb:00b1ab\n"
	"ec flow2 rgb:ddd\n"
	"ec fname rgb:00ffff\n"
	"ec label rgb:66d9ef\n"
	"ec math rgb:ffffff\n"
	"ec mov rgb:ffffff\n"
	"ec nop rgb:FFE792\n"
	"ec offset rgb:ababab\n"
	"ec prompt rgb:fc0\n"
	"ec pop rgb:ef8d1a\n"
	"ec push rgb:df077e\n"
	"ec reg rgb:009d9d\n"
	"ec args rgb:ffd200\n"
	"ec help rgb:009d9d\n"
	"ec input rgb:888\n"
	"ec trap rgb:d66\n"
	"ec ucall rgb:ffd200\n"
	"ec ujmp rgb:9dd600\n"
	"ec swi rgb:df077e\n"
	"ec creg rgb:69d200\n"
	"ec flag rgb:00b1ab\n"
	"ec linehl rgb:00007f\n"
	"ec ai.read rgb:66d\n"
	"ec ai.write rgb:d66\n"
	"ec ai.exec rgb:6d6\n"
	"ec graph.box rgb:5202f0\n"
	"ec graph.box2 rgb:ffcc00\n"
	"ec graph.box3 rgb:66d9ef\n"
	"ec graph.box4 rgb:66d9ef\n"
	"ec graph.true rgb:009100\n"
	"ec graph.false rgb:bc0000\n"
	"ec graph.trufae rgb:0043cb\n"
	"ec graph.current rgb:fff\n"
	"\n"
	"ec var rgb:99a\n"
	"ec var.type rgb:df077e\n"
	"ec var.addr rgb:ababab\n"
	"ec var.name rgb:99a\n"
	"\n"
	"ec ai.read rgb:be077e\n"
	"ec ai.write rgb:df077e\n"
	"ec ai.exec rgb:9dd600\n"
	"ec ai.seq rgb:9dd600\n"
	"ec ai.ascii rgb:d25032\n"
	"\n"
	"ec widget.bg rgb:099 rgb:444\n"
	"ec widget.sel rgb:000 yellow\n"
;
static const char * const theme_focus = \
	"ecd\n"
	"ec fname rgb:0cf rgb:000\n"
	"ec label rgb:0cf rgb:000\n"
	"ec math rgb:ccc\n"
	"ec bin rgb:f90\n"
	"ec call rgb:ff0 rgb:00f\n"
	"ec ucall rgb:000 rgb:ff0\n"
	"ec jmp rgb:ff0 rgb:00f\n"
	"ec cjmp rgb:ff0\n"
	"ec ujmp rgb:ff3\n"
	"ec ret rgb:f00\n"
	"ec trap rgb:000 rgb:f00\n"
	"ec offset rgb:fff\n"
	"ec prompt byellow\n"
	"ec comment rgb:0cf\n"
	"ec usrcmt rgb:ff0 rgb:06f\n"
	"ec push rgb:f3f\n"
	"ec pop rgb:f3f bold\n"
	"ec cmp rgb:ff0 rgb:060\n"
	"ec nop rgb:000\n"
	"ec b0x00 green\n"
	"ec b0x7f cyan\n"
	"ec b0xff red\n"
	"ec btext yellow\n"
	"ec other rgb:666\n"
	"ec num rgb:ff0\n"
	"ec reg rgb:6f0 rgb:000\n"
	"ec fline rgb:fc0\n"
	"ec flow rgb:ff3\n"
	"ec flow2 rgb:fff\n"
	"ec mov rgb:888\n"
	"ec swi rgb:000 rgb:fff\n"
	"ec floc rgb:af2\n"
	"ec linehl rgb:222\n"
	"ec ai.read green\n"
	"ec ai.write blue\n"
	"ec ai.exec red\n"
	"ec ai.seq rgb:ff00ff\n"
	"ec ai.ascii rgb:ffff00\n"
	"\n"
	"ec graph.box blue\n"
	"ec graph.box2 yellow\n"
	"ec graph.box3 blue\n"
	"ec graph.box4 blue\n"
	"ec graph.true green\n"
	"ec graph.false red\n"
	"ec graph.trufae blue\n"
	"ec graph.current blue\n"
	"ec graph.traced red\n"
	"\n"
	"ec var cyan\n"
	"ec var.type rgb:f3f\n"
	"ec var.addr white\n"
	"ec var.name cyan\n"
	"\n"
	"ec ai.read rgb:8f0\n"
	"ec ai.write rgb:6f0\n"
	"ec ai.exec rgb:f3f\n"
	"ec ai.seq white\n"
	"ec ai.ascii yellow\n"
	"\n"
	"ec widget.bg rgb:ff0 rgb:000\n"
	"ec widget.sel rgb:000 rgb:0f0\n"
;
static const char * const theme_gb = \
	"ecd\n"
	"ec comment rgb:86c06c\n"
	"ec usrcmt rgb:86c06c\n"
	"ec mov rgb:a0c090\n"
	"ec fname rgb:e0f8cf\n"
	"ec args white\n"
	"ec floc rgb:86c06c\n"
	"ec fline rgb:e0f8cf\n"
	"ec flag rgb:e0f8cf\n"
	"ec invalid rgb:306850\n"
	"ec input rgb:e0f8cf\n"
	"ec label rgb:86c06c\n"
	"ec flow rgb:86c06c\n"
	"ec flow2 rgb:86c06c\n"
	"ec prompt rgb:e0f8cf\n"
	"ec offset rgb:306850\n"
	"ec other rgb:306850\n"
	"ec b0x00 rgb:306850\n"
	"ec b0x7f rgb:86c06c\n"
	"ec b0xff rgb:e0f8cf\n"
	"ec math rgb:86c06c\n"
	"ec crypto rgb:86c06c\n"
	"ec bin rgb:86c06c\n"
	"ec btext rgb:86c06c\n"
	"ec push rgb:86c06c\n"
	"ec pop rgb:86c06c\n"
	"ec jmp rgb:e0f8cf\n"
	"ec cjmp rgb:e0f8cf\n"
	"ec ujmp rgb:e0f8cf\n"
	"ec call rgb:e0f8cf\n"
	"ec trap rgb:e0f8cf\n"
	"ec swi rgb:e0f8cf\n"
	"ec ret rgb:e0f8cf\n"
	"ec cmp rgb:86c06c\n"
	"ec nop rgb:306850\n"
	"ec trap rgb:86c06c\n"
	"ec reg rgb:e0f8cf\n"
	"ec creg rgb:e8f8cf\n"
	"ec num rgb:e0f8cf\n"
	"ec help rgb:e0f8cf\n"
	"ec trap rgb:000 rgb:86c06c\n"
	"ec ucall rgb:86c06c\n"
	"ec linehl rgb:306850\n"
	"\n"
	"ec graph.box rgb:86c06c\n"
	"ec graph.box2 rgb:86c06c\n"
	"ec graph.box3 rgb:86c06c\n"
	"ec graph.box4 rgb:86c06c\n"
	"ec graph.true rgb:86c06c\n"
	"ec graph.false rgb:e0f8cf\n"
	"ec graph.trufae rgb:86c06c\n"
	"ec graph.current rgb:e0f8cf\n"
	"ec graph.traced rgb:86c06c\n"
	"\n"
	"ec gui.cflow rgb:e0f8cf\n"
	"ec gui.dataoffset rgb:e0f8cf\n"
	"ec gui.background rgb:e0f8cf\n"
	"ec gui.background2 rgb:e0f8cf\n"
	"ec gui.border rgb:e0f8cf\n"
	"\n"
	"ec ai.read rgb:e0f8cf\n"
	"ec ai.write rgb:86c06c\n"
	"ec ai.exec rgb:86c06c\n"
	"ec ai.seq rgb:86c06c\n"
	"ec ai.ascii rgb:86c06c\n"
	"\n"
	"ec var rgb:86c06c\n"
	"ec var.type rgb:86c06c\n"
	"ec var.addr rgb:e0f8cf\n"
	"ec var.name rgb:86c06c\n"
	"\n"
	"ec widget.bg rgb:86c06c rgb:000\n"
	"ec widget.sel rgb:e0f8cf rgb:000\n"
;
static const char * const theme_gentoo = \
	"ecd\n"
	"ec comment cyan\n"
	"ec usrcmt cyan\n"
	"ec args cyan\n"
	"ec fname green\n"
	"ec floc yellow\n"
	"ec fline blue\n"
	"ec flag green\n"
	"ec label yellow\n"
	"ec help cyan\n"
	"ec flow cyan\n"
	"ec flow2 blue\n"
	"ec prompt green\n"
	"ec offset yellow\n"
	"ec input rgb:f03030\n"
	"ec invalid red\n"
	"ec other rgb:c0c0c0\n"
	"ec b0x00 rgb:808080\n"
	"ec b0x7f rgb:c0c0c0\n"
	"ec b0xff cyan\n"
	"ec math white\n"
	"ec bin rgb:f0f0f0\n"
	"ec btext rgb:009000\n"
	"ec push yellow\n"
	"ec pop yellow\n"
	"ec crypto cyan\n"
	"ec call green\n"
	"ec ucall green\n"
	"ec nop rgb:3030f0\n"
	"ec ret rgb:30f030\n"
	"ec trap rgb:f03030\n"
	"ec ujmp rgb:009000\n"
	"ec jmp rgb:009000\n"
	"ec cjmp rgb:009000\n"
	"ec swi black yellow\n"
	"ec cmp cyan\n"
	"ec reg rgb:f0f0f0\n"
	"ec creg yellow\n"
	"ec num yellow black\n"
	"ec mov rgb:d0d0d0\n"
	"ec var white\n"
	"ec var.type yellow\n"
	"ec var.addr white\n"
	"ec var.name white\n"
	"ec ai.read rgb:e0e0e0\n"
	"ec ai.write rgb:d0d0d0\n"
	"ec ai.exec rgb:f0f0f0\n"
	"ec ai.seq rgb:f03030\n"
	"ec ai.ascii rgb:f03030\n"
	"ec graph.box rgb:3030f0\n"
	"ec graph.box2 cyan\n"
	"ec graph.box3 rgb:30f030\n"
	"ec graph.box4 rgb:30f030\n"
	"ec graph.true rgb:30f030\n"
	"ec graph.false rgb:f03030\n"
	"ec graph.trufae rgb:3030f0\n"
	"ec graph.current rgb:f03030\n"
	"ec graph.traced rgb:f00000\n"
	"ec gui.cflow rgb:f03030\n"
	"ec gui.dataoffset rgb:3030f0\n"
	"ec gui.background rgb:000000\n"
	"ec gui.background2 rgb:ffffff\n"
	"ec gui.border rgb:000000\n"
	"ec wordhl rgb:ff0000\n"
	"ec linehl rgb:101010\n"
	"\n"
	"ec widget.bg cyan rgb:3030f0\n"
	"ec widget.sel yellow rgb:090\n"
;
static const char * const theme_greepy = \
	"ecd\n"
	"ec b0x00 rgb:ccc\n"
	"ec b0x7f rgb:fff\n"
	"ec b0xff rgb:fff\n"
	"ec args rgb:fff\n"
	"ec bin rgb:fff\n"
	"ec btext rgb:fff\n"
	"ec call rgb:f0f\n"
	"ec cjmp rgb:f0f\n"
	"ec cmp rgb:aaa\n"
	"ec comment rgb:0f0\n"
	"ec usrcmt rgb:0f0\n"
	"ec reg rgb:fff\n"
	"ec creg rgb:f0f\n"
	"ec flag rgb:f0f\n"
	"ec fline rgb:aaa\n"
	"ec flow rgb:f0f\n"
	"ec flow2 rgb:fff\n"
	"ec fname rgb:f0f\n"
	"ec help rgb:f0f\n"
	"ec input rgb:f0f\n"
	"ec jmp rgb:f0f\n"
	"ec label rgb:f0f\n"
	"ec math rgb:fff\n"
	"ec mov rgb:fff\n"
	"ec nop rgb:fff\n"
	"ec num rgb:0f0\n"
	"ec offset rgb:0f0\n"
	"ec other rgb:999\n"
	"ec pop rgb:999\n"
	"ec prompt rgb:f0f\n"
	"ec push rgb:0f0\n"
	"ec ret rgb:fff\n"
	"ec swi rgb:f0f\n"
	"ec trap rgb:f0f\n"
	"ec ucall rgb:f0f\n"
	"ec ujmp rgb:f0f\n"
	"ec graph.true rgb:f0f\n"
	"ec graph.false rgb:fff\n"
	"ec graph.trufae rgb:0f0\n"
	"ec graph.current rgb:f0f\n"
	"ec graph.traced rgb:f00\n"
	"ec graph.box rgb:fff\n"
	"ec graph.box2 rgb:f0f\n"
	"ec graph.box3 rgb:000\n"
	"ec graph.box4 rgb:000\n"
	"ec gui.cflow rgb:f0f\n"
	"ec gui.dataoffset rgb:f0f\n"
	"ec crypto rgb:f0f\n"
	"ec floc rgb:fff\n"
	"ec linehl rgb:111\n"
	"\n"
	"ec var rgb:f0f\n"
	"ec var.type rgb:fff\n"
	"ec var.addr rgb:0f0\n"
	"ec var.name rgb:f0f\n"
	"\n"
	"ec ai.read rgb:eee\n"
	"ec ai.write rgb:0f0\n"
	"ec ai.exec rgb:fff\n"
	"ec ai.seq rgb:f0f\n"
	"ec ai.ascii rgb:f0f\n"
	"\n"
	"ec widget.bg rgb:f0f rgb:0f0\n"
	"ec widget.sel rgb:0f0 rgb:f0f\n"
;
static const char * const theme_gruvbox = \
	"# Inspired by gruvbox\n"
	"#https://github.com/morhetz/gruvbox\n"
	"ecd\n"
	"ec ai.exec rgb:689d6a\n"
	"ec ai.read rgb:458588\n"
	"ec ai.write rgb:cc241d\n"
	"ec args rgb:b8bb26\n"
	"ec b0x00 rgb:665c54\n"
	"ec b0x7f rgb:504945\n"
	"ec b0xff rgb:ebdbb2\n"
	"ec btext rgb:8ec07c\n"
	"ec comment rgb:928374\n"
	"ec creg rgb:fe8019\n"
	"ec flag rgb:8ec07c\n"
	"ec fline rgb:83a598\n"
	"ec flow rgb:928374\n"
	"ec flow2 rgb:ebdbb2\n"
	"ec fname rgb:8ec07c\n"
	"ec graph.traced rgb:b8bb26\n"
	"ec graph.box rgb:3c3836\n"
	"ec graph.box2 rgb:fe8019\n"
	"ec graph.box3 rgb:83a598\n"
	"ec graph.box4 rgb:83a598\n"
	"ec graph.current rgb:ebdbb2\n"
	"ec graph.false rgb:fb4934\n"
	"ec graph.true rgb:b8bb26\n"
	"ec graph.trufae rgb:83a598\n"
	"ec help rgb:b8bb26\n"
	"ec input rgb:7c6f64\n"
	"ec jmp rgb:b8bb26\n"
	"ec label rgb:8ec07c\n"
	"ec linehl rgb:3c3836\n"
	"ec mov rgb:ebdbb2\n"
	"ec nop rgb:928374\n"
	"ec num rgb:d3869b\n"
	"ec offset rgb:d65d0e\n"
	"ec other rgb:7c6f64\n"
	"ec prompt rgb:8ec07c\n"
	"ec call rgb:fb4934\n"
	"ec reg rgb:fabd2f\n"
	"ec ret rgb:cc241d\n"
	"ec swi rgb:cc241d\n"
	"ec trap rgb:cc241d\n"
	"ec cmp rgb:8ec07c\n"
	"ec math rgb:665c54\n"
	"ec bin rgb:665c54\n"
	"ec push rgb:8ec07c\n"
	"ec pop rgb:8ec07c\n"
	"ec cjmp rgb:b8bb26\n"
	"ec ucall rgb:83a598\n"
	"ec ujmp rgb:b8bb26\n"
	"ec usrcmt rgb:fe8019\n"
	"\n"
	"ec var rgb:83a598\n"
	"ec var.type rgb:fabd2f\n"
	"ec var.addr rgb:fb4934\n"
	"ec var.name rgb:83a598\n"
	"\n"
	"ec widget.bg rgb:7c6f64 rgb:282828\n"
	"ec widget.sel rgb:8ec07c rgb:3c3836\n"
;
static const char * const theme_iaito = \
	"ecd\n"
	"ec comment rgb:8e8c8b\n"
	"ec usrcmt rgb:8e8c8b\n"
	"ec args rgb:20b0c0\n"
	"ec fname rgb:0693c9\n"
	"ec floc rgb:00ffff\n"
	"ec fline rgb:5f8700\n"
	"ec flag rgb:0a8ab0\n"
	"ec label rgb:0068bd\n"
	"ec help rgb:005f87\n"
	"ec flow rgb:5f8700\n"
	"ec flow2 rgb:5f8700\n"
	"ec prompt rgb:30a000\n"
	"ec offset rgb:8e8c8b\n"
	"ec input rgb:000000\n"
	"ec invalid rgb:e03030\n"
	"ec other rgb:005f87\n"
	"ec b0x00 rgb:707070\n"
	"ec b0x7f rgb:7f7f7f\n"
	"ec b0xff rgb:e03030\n"
	"ec math rgb:005f87\n"
	"ec bin rgb:005f87\n"
	"ec btext rgb:30a000\n"
	"ec push rgb:da3192\n"
	"ec pop rgb:da3192\n"
	"ec crypto rgb:d03080\n"
	"ec jmp rgb:5f8700\n"
	"ec cjmp rgb:5f8700\n"
	"ec ujmp rgb:5f8700\n"
	"ec call rgb:0693c9\n"
	"ec ucall rgb:0693c9\n"
	"ec nop rgb:005f87\n"
	"ec ret rgb:0693c9\n"
	"ec trap rgb:0693c9\n"
	"ec swi rgb:5f8700\n"
	"ec cmp rgb:da3192\n"
	"ec reg rgb:005f87\n"
	"ec creg rgb:005f87\n"
	"ec num rgb:ef5919\n"
	"ec mov rgb:da3192\n"
	"ec var rgb:d03080\n"
	"ec var.type rgb:876ac1\n"
	"ec var.addr rgb:b0b0b0\n"
	"ec var.name rgb:d03080\n"
	"ec widget.bg rgb:303030\n"
	"ec widget.sel rgb:ff0000\n"
	"ec ai.read rgb:1080d0\n"
	"ec ai.write rgb:005f87\n"
	"ec ai.exec rgb:b08000\n"
	"ec ai.seq rgb:d03080\n"
	"ec ai.ascii rgb:000000\n"
	"ec graph.box rgb:005f87\n"
	"ec graph.box2 rgb:5f8700\n"
	"ec graph.box3 rgb:e03030\n"
	"ec graph.box4 rgb:005f87\n"
	"ec graph.true rgb:5f8700\n"
	"ec graph.false rgb:e03030\n"
	"ec graph.trufae rgb:005f87\n"
	"ec graph.current rgb:00f0f0\n"
	"ec graph.traced rgb:e03030\n"
	"ec gui.cflow rgb:ffff00\n"
	"ec gui.dataoffset rgb:ffff00\n"
	"ec gui.background rgb:000000\n"
	"ec gui.background2 rgb:ffffff\n"
	"ec gui.border rgb:000000\n"
	"ec wordhl rgb:ff0000\n"
	"ec linehl rgb:efefef\n"
	"\n"
	"ec widget.bg rgb:4cf rgb:058\n"
	"ec widget.sel black rgb:d39\n"
;
static const char * const theme_lima = \
	"ecd\n"
	"ec comment rgb:af2\n"
	"ec usrcmt rgb:af2 rgb:140\n"
	"ec mov rgb:090\n"
	"ec fname rgb:27a\n"
	"ec args white\n"
	"ec floc rgb:af2\n"
	"ec fline rgb:af2\n"
	"ec flag rgb:af2\n"
	"ec fname rgb:af2\n"
	"ec label rgb:af2\n"
	"ec flow rgb:af2\n"
	"ec flow2 rgb:af2\n"
	"ec prompt rgb:090\n"
	"ec offset rgb:ef0\n"
	"ec input rgb:7d5\n"
	"ec other rgb:090\n"
	"ec b0x00 rgb:040\n"
	"ec b0x7f rgb:090\n"
	"ec b0xff rgb:af2\n"
	"ec math rgb:090\n"
	"ec crypto rgb:090\n"
	"ec bin rgb:ef0\n"
	"ec btext rgb:ef0\n"
	"ec push rgb:ef0\n"
	"ec pop rgb:ef0\n"
	"ec jmp rgb:af2\n"
	"ec cjmp rgb:af2\n"
	"ec ujmp rgb:af2\n"
	"ec call rgb:fff\n"
	"ec ucall rgb:fff\n"
	"ec nop rgb:2da\n"
	"ec ret rgb:5d5\n"
	"ec trap rgb:a7a\n"
	"ec swi rgb:dd5\n"
	"ec cmp rgb:0a0\n"
	"ec reg rgb:af2\n"
	"ec creg rgb:af2\n"
	"ec num rgb:ef0\n"
	"ec help rgb:af2\n"
	"ec invalid rgb:ef0\n"
	"ec trap rgb:000 rgb:af2\n"
	"ec linehl rgb:140\n"
	"\n"
	"ec graph.box rgb:af2\n"
	"ec graph.box2 rgb:0a0\n"
	"ec graph.box3 rgb:444\n"
	"ec graph.box4 rgb:444\n"
	"ec graph.true rgb:ef2\n"
	"ec graph.false rgb:5d5\n"
	"ec graph.trufae rgb:af2\n"
	"ec graph.current rgb:af2\n"
	"ec graph.traced rgb:090\n"
	"\n"
	"ec var yellow\n"
	"ec var.type green\n"
	"ec var.addr yellow\n"
	"ec var.name yellow\n"
	"\n"
	"ec ai.read rgb:8e2\n"
	"ec ai.write rgb:9e2\n"
	"ec ai.exec rgb:af2\n"
	"ec ai.seq rgb:ef0\n"
	"ec ai.ascii rgb:ef0\n"
	"\n"
	"ec widget.bg rgb:ef0 rgb:060\n"
	"ec widget.sel black rgb:af2\n"
;
static const char * const theme_matrix = \
	"ecd\n"
	"ec b0x00 rgb:030\n"
	"ec b0x7f rgb:0f0\n"
	"ec b0xff rgb:050\n"
	"ec bin rgb:0f0\n"
	"ec crypto rgb:0f0\n"
	"ec other rgb:0a0\n"
	"ec btext rgb:0b0\n"
	"ec ucall rgb:0b0\n"
	"ec call rgb:0b0\n"
	"ec ujmp rgb:2f2\n"
	"ec cjmp rgb:2f2\n"
	"ec jmp rgb:2f2\n"
	"ec cmp rgb:191\n"
	"ec comment rgb:060\n"
	"ec usrcmt rgb:3f3\n"
	"ec ret rgb:0b0\n"
	"ec fline rgb:0a0\n"
	"ec flow bgreen\n"
	"ec flow2 green\n"
	"ec fname rgb:0b0\n"
	"ec label rgb:0f0\n"
	"ec math rgb:040\n"
	"ec invalid rgb:060\n"
	"ec mov rgb:383\n"
	"ec nop rgb:030\n"
	"ec num rgb:3f3\n"
	"ec offset rgb:383\n"
	"ec pop rgb:040\n"
	"ec prompt rgb:080\n"
	"ec push rgb:060\n"
	"ec reg rgb:1a1\n"
	"ec args rgb:0b0\n"
	"ec help rgb:0b0\n"
	"ec input rgb:0a0\n"
	"ec trap rgb:040\n"
	"ec swi rgb:0b0\n"
	"ec creg rgb:080\n"
	"ec flag rgb:0f0\n"
	"ec linehl rgb:020\n"
	"ec graph.box green\n"
	"ec graph.box2 rgb:060\n"
	"ec graph.box3 rgb:ff0000\n"
	"ec graph.box4 rgb:0000ff\n"
	"ec graph.true green\n"
	"ec graph.false rgb:050\n"
	"ec graph.trufae green\n"
	"ec graph.current green\n"
	"ec graph.traced rgb:060\n"
	"\n"
	"ec var rgb:060\n"
	"ec var.type rgb:383\n"
	"ec var.addr rgb:060\n"
	"ec var.name rgb:060\n"
	"\n"
	"ec ai.read rgb:0a1\n"
	"ec ai.write rgb:1a1\n"
	"ec ai.exec rgb:383\n"
	"ec ai.seq rgb:3f3\n"
	"ec ai.ascii rgb:3f3\n"
	"\n"
	"ec widget.bg rgb:0b0 rgb:000\n"
	"ec widget.sel black rgb:0c0\n"
;
static const char * const theme_monokai = \
	"ecd\n"
	"ec b0x00 rgb:555\n"
	"ec b0x7f rgb:fff\n"
	"ec b0xff rgb:eee\n"
	"ec other rgb:888\n"
	"ec bin rgb:FFE792\n"
	"ec btext rgb:a398e5\n"
	"ec ucall rgb:f1c40f\n"
	"ec call rgb:f1c40f\n"
	"ec ujmp rgb:A6df2E\n"
	"ec cjmp rgb:A6df2E\n"
	"ec jmp rgb:A6df2E\n"
	"ec cmp rgb:ef8d1a\n"
	"ec comment rgb:66d9ef\n"
	"ec usrcmt rgb:f92672\n"
	"ec ret rgb:d66\n"
	"ec num rgb:a398e5\n"
	"ec fline rgb:2bc\n"
	"ec flow rgb:A6df2E\n"
	"ec flow2 rgb:ddd\n"
	"ec fname rgb:66d9ef\n"
	"ec label rgb:66d9ef\n"
	"ec math rgb:ddd\n"
	"ec mov rgb:ddd\n"
	"ec nop rgb:FFE792\n"
	"ec offset rgb:f92672\n"
	"ec prompt rgb:ef8d1a\n"
	"ec pop rgb:ef8d1a\n"
	"ec push rgb:ef8d1a\n"
	"ec reg rgb:fff\n"
	"ec args rgb:2bc\n"
	"ec help rgb:ef8d1a\n"
	"ec input rgb:888\n"
	"ec trap rgb:d66\n"
	"ec swi rgb:d66\n"
	"ec creg rgb:df7\n"
	"ec flag rgb:ef8d1a\n"
	"ec linehl rgb:008\n"
	"ec ai.read rgb:66d\n"
	"ec ai.write rgb:d66\n"
	"ec ai.exec rgb:6d6\n"
	"ec ai.ascii rgb:a398e5\n"
	"ec graph.box rgb:66d9ef\n"
	"ec graph.box2 rgb:ef8d1a\n"
	"ec graph.box3 rgb:66d9ef\n"
	"ec graph.box4 rgb:66d9ef\n"
	"ec graph.true rgb:A6df2E\n"
	"ec graph.false rgb:f92672\n"
	"ec graph.trufae rgb:a398e5\n"
	"ec graph.current rgb:fff\n"
	"\n"
	"ec var cyan\n"
	"ec var.type rgb:a398e5\n"
	"ec var.addr rgb:f92672\n"
	"ec var.name cyan\n"
	"\n"
	"ec widget.bg rgb:f1c40f rgb:555\n"
	"ec widget.sel white rgb:f92672\n"
;
static const char * const theme_ogray = \
	"ecd\n"
	"ec b0x00 rgb:ccc\n"
	"ec b0x7f rgb:777\n"
	"ec b0xff rgb:333\n"
	"ec args rgb:fff\n"
	"ec bin rgb:fff\n"
	"ec btext rgb:fff\n"
	"ec call rgb:f72 rgb:333\n"
	"ec cjmp rgb:f72\n"
	"ec cmp rgb:aaa\n"
	"ec comment rgb:f72\n"
	"ec usrcmt rgb:000 rgb:f72\n"
	"ec reg rgb:fff\n"
	"ec creg rgb:f72\n"
	"ec flag rgb:f72\n"
	"ec fline rgb:aaa\n"
	"ec flow rgb:f72\n"
	"ec flow2 rgb:fff\n"
	"ec fname rgb:f72\n"
	"ec help rgb:f72\n"
	"ec input rgb:f72\n"
	"ec jmp rgb:f72\n"
	"ec label rgb:f72\n"
	"ec math rgb:777\n"
	"ec mov rgb:ddd\n"
	"ec nop rgb:333\n"
	"ec num rgb:f72\n"
	"ec offset rgb:555\n"
	"ec other rgb:999\n"
	"ec pop rgb:999\n"
	"ec prompt rgb:f72\n"
	"ec push rgb:ddd\n"
	"ec ret rgb:777\n"
	"ec swi rgb:f72\n"
	"ec trap rgb:f72\n"
	"ec ucall rgb:f72 rgb:333\n"
	"ec ujmp rgb:f72\n"
	"ec graph.true rgb:f72\n"
	"ec graph.false rgb:777\n"
	"ec graph.trufae rgb:ddd\n"
	"ec graph.current rgb:f72\n"
	"ec graph.traced rgb:f00\n"
	"ec graph.box rgb:fff\n"
	"ec graph.box2 rgb:f72\n"
	"ec graph.box3 rgb:000\n"
	"ec graph.box4 rgb:000\n"
	"ec gui.cflow rgb:f72\n"
	"ec gui.dataoffset rgb:f72\n"
	"ec crypto rgb:f72\n"
	"ec floc rgb:777\n"
	"ec linehl rgb:111\n"
	"\n"
	"ec var rgb:f72\n"
	"ec var.type rgb:777\n"
	"ec var.addr rgb:555\n"
	"ec var.name rgb:f72\n"
	"\n"
	"ec ai.read rgb:eee\n"
	"ec ai.write rgb:ddd\n"
	"ec ai.exec rgb:fff\n"
	"ec ai.seq rgb:f72\n"
	"ec ai.ascii rgb:f72\n"
	"\n"
	"ec widget.bg rgb:f72 rgb:333\n"
	"ec widget.sel rgb:333 rgb:f72\n"
;
static const char * const theme_onedark = \
	"#Inspired by onedark.vim\n"
	"# e scr.color = 3 // to enable 16m color\n"
	"ecd\n"
	"ec var rgb:61afef\n"
	"ec var.type white\n"
	"ec var.addr rgb:4b5263\n"
	"ec var.name rgb:61afef\n"
	"ec b0x00 rgb:c678dd\n"
	"ec b0xff rgb:be5046\n"
	"ec prompt rgb:d19a66\n"
	"ec fname rgb:56b6c2\n"
	"ec label rgb:0f3\n"
	"ec math rgb:660\n"
	"ec bin rgb:e5c07b\n"
	"ec btext rgb:e5c07b\n"
	"ec call rgb:e06c75\n"
	"ec ucall rgb:e06c75\n"
	"ec jmp rgb:c678dd\n"
	"ec cjmp rgb:c678dd\n"
	"ec ujmp rgb:c678dd\n"
	"ec offset rgb:366\n"
	"ec comment rgb:56b6c2\n"
	"ec usrcmt rgb:ff0\n"
	"ec push rgb:61afef\n"
	"ec pop rgb:61afef\n"
	"ec cmp rgb:060\n"
	"ec nop rgb:abb2bf\n"
	"ec num rgb:be5046\n"
	"ec reg rgb:e5c07b\n"
	"ec fline rgb:fc0\n"
	"ec flow rgb:98c379\n"
	"ec flow2 rgb:98c379\n"
	"ec args rgb:d19a66\n"
	"ec floc rgb:fc0\n"
	"ec flag rgb:56b6c2\n"
	"ec label rgb:56b6c2\n"
	"ec help rgb:98c379\n"
	"ec input white\n"
	"ec invalid rgb:e06c75\n"
	"ec crypto rgb:0000ff\n"
	"ec ret rgb:be5046\n"
	"ec trap rgb:e06c75\n"
	"ec swi rgb:e06c75\n"
	"ec mov rgb:98c379\n"
	"ec creg rgb:e5c07b\n"
	"ec ai.read rgb:98c379\n"
	"ec ai.write rgb:61afef\n"
	"ec ai.exec rgb:e06c75\n"
	"ec ai.seq rgb:ff00ff\n"
	"ec ai.ascii rgb:ffff00\n"
	"ec gui.cflow rgb:d19a66\n"
	"ec gui.dataoffset rgb:d19a66\n"
	"ec gui.background black\n"
	"ec gui.background2 white\n"
	"ec gui.border black\n"
	"ec linehl rgb:61afef\n"
	"ec graph.box rgb:d19a66\n"
	"ec graph.box2 rgb:61afef\n"
	"ec graph.box3 rgb:c678dd\n"
	"ec graph.box4 rgb:000000\n"
	"ec graph.true rgb:98c379\n"
	"ec graph.false rgb:e06c75\n"
	"ec graph.trufae rgb:61afef\n"
	"ec graph.current rgb:61afef\n"
	"ec graph.traced rgb:d19a66\n"
	"ec linehl rgb:5c6370\n"
	"\n"
	"ec widget.bg rgb:ec7 rgb:366\n"
	"ec widget.sel black rgb:98c379\n"
;
static const char * const theme_pink = \
	"ecd\n"
	"ec comment rgb:72d\n"
	"ec help rgb:f5d\n"
	"ec fname rgb:f5d\n"
	"ec fline rgb:afd\n"
	"ec flag rgb:f5d\n"
	"ec label rgb:f5d\n"
	"ec flow rgb:f5d\n"
	"ec flow2 rgb:72d\n"
	"ec prompt rgb:f5d\n"
	"ec offset rgb:72d\n"
	"ec input rgb:f5d\n"
	"ec other rgb:d5a\n"
	"ec b0x00 rgb:a27\n"
	"ec b0x7f rgb:55d\n"
	"ec b0xff rgb:f5d\n"
	"ec mov rgb:fff\n"
	"ec bin rgb:fff\n"
	"ec math rgb:fff\n"
	"ec btext rgb:72d\n"
	"ec push rgb:75a\n"
	"ec pop rgb:75a\n"
	"ec jmp rgb:f5d\n"
	"ec cjmp rgb:f5d\n"
	"ec ujmp rgb:f5d\n"
	"ec usrcmt rgb:f5d rgb:502\n"
	"ec call rgb:000 rgb:faf\n"
	"ec ucall rgb:000 rgb:faf\n"
	"ec nop rgb:fa5\n"
	"ec ret rgb:ffa\n"
	"ec trap rgb:a22\n"
	"ec swi rgb:255\n"
	"ec cmp rgb:a7f\n"
	"ec reg rgb:f5d\n"
	"ec creg rgb:f5d\n"
	"ec num rgb:72d\n"
	"ec crypto rgb:75d\n"
	"ec linehl rgb:fdf\n"
	"ec graph.true rgb:f5d\n"
	"ec graph.false rgb:715\n"
	"ec graph.trufae rgb:f5d\n"
	"ec graph.current rgb:f5d\n"
	"ec graph.traced rgb:a7f\n"
	"ec graph.box rgb:f5d\n"
	"ec graph.box2 rgb:f5d\n"
	"ec graph.box3 rgb:f5d\n"
	"ec graph.box4 rgb:f5d\n"
	"ec gui.cflow rgb:f5d\n"
	"ec gui.dataoffset rgb:f5d\n"
	"ec gui.background black\n"
	"ec gui.background2 magenta\n"
	"## gui.border\n"
	"\n"
	"ec var rgb:72d\n"
	"ec var.type blue\n"
	"ec var.addr rgb:72d\n"
	"ec var.name rgb:72d\n"
	"\n"
	"ec ai.read rgb:f6e\n"
	"ec ai.write rgb:f5d\n"
	"ec ai.exec rgb:75a\n"
	"ec ai.seq rgb:72d\n"
	"ec ai.ascii rgb:72d\n"
	"\n"
	"ec widget.bg rgb:72d rgb:222\n"
	"ec widget.sel white rgb:f5d\n"
;
static const char * const theme_rasta = \
	"ecd\n"
	"ec comment green\n"
	"ec usrcmt yellow\n"
	"ec fname yellow\n"
	"ec fline green\n"
	"ec flag green\n"
	"ec label yellow\n"
	"ec flow green\n"
	"ec flow2 yellow\n"
	"ec floc green\n"
	"ec prompt green\n"
	"ec offset yellow\n"
	"ec input green\n"
	"ec other green\n"
	"ec b0x00 yellow\n"
	"ec b0x7f green\n"
	"ec b0xff red\n"
	"ec math yellow\n"
	"ec bin green\n"
	"ec btext red\n"
	"ec push yellow\n"
	"ec pop green\n"
	"ec jmp red\n"
	"ec cjmp red\n"
	"ec ujmp red\n"
	"ec call red\n"
	"ec ucall red\n"
	"ec nop red\n"
	"ec ret yellow\n"
	"ec trap green\n"
	"ec swi red\n"
	"ec cmp yellow\n"
	"ec reg green\n"
	"ec num red\n"
	"ec mov yellow\n"
	"ec crypto yellow\n"
	"ec args green\n"
	"ec help yellow\n"
	"ec input green\n"
	"ec trap red\n"
	"ec swi red\n"
	"ec creg green\n"
	"ec flag yellow\n"
	"ec linehl rgb:008\n"
	"ec graph.box yellow\n"
	"ec graph.box2 green\n"
	"ec graph.box3 red\n"
	"ec graph.box4 red\n"
	"ec graph.true green\n"
	"ec graph.false red\n"
	"ec graph.trufae green\n"
	"ec graph.current green\n"
	"ec graph.traced yellow\n"
	"\n"
	"ec var green\n"
	"ec var.type red\n"
	"ec var.addr yellow\n"
	"ec var.name green\n"
	"\n"
	"ec ai.read green\n"
	"ec ai.write green\n"
	"ec ai.exec yellow\n"
	"ec ai.seq red\n"
	"ec ai.ascii red\n"
	"\n"
	"ec widget.bg yellow black\n"
	"ec widget.sel black green\n"
;
static const char * const theme_sepia = \
	"# requires e scr.color=3\n"
	"ecd\n"
	"ec b0x00 rgb:431\n"
	"ec b0x7f rgb:fd9\n"
	"ec b0xff rgb:960\n"
	"ec bin rgb:fec\n"
	"ec crypto rgb:0000ff\n"
	"ec other rgb:ba6\n"
	"ec btext rgb:fec\n"
	"ec ucall rgb:ffffff\n"
	"ec call rgb:ffffff\n"
	"ec ujmp rgb:fec\n"
	"ec cjmp rgb:fec\n"
	"ec jmp rgb:fec\n"
	"ec cmp rgb:ca6\n"
	"ec comment rgb:fd9\n"
	"ec usrcmt rgb:fff\n"
	"ec ret rgb:ab7\n"
	"ec fline rgb:ca6\n"
	"ec floc rgb:00ffff\n"
	"ec flow rgb:ca6\n"
	"ec flow2 rgb:fec\n"
	"ec fname rgb:bb7\n"
	"ec label rgb:788\n"
	"ec math rgb:db7\n"
	"ec invalid rgb:c64\n"
	"ec mov rgb:db7\n"
	"ec nop rgb:788\n"
	"ec num rgb:960\n"
	"ec offset rgb:960\n"
	"ec pop rgb:bb7\n"
	"ec prompt rgb:960\n"
	"ec push rgb:db7\n"
	"ec reg rgb:ca6\n"
	"ec args rgb:bb7\n"
	"ec help rgb:ab7\n"
	"ec input rgb:ca6\n"
	"ec trap rgb:431\n"
	"ec swi rgb:bb7\n"
	"ec creg rgb:960\n"
	"ec flag rgb:ffd\n"
	"ec wordhl rgb:ff0000\n"
	"ec linehl rgb:210\n"
	"ec graph.true rgb:fd9\n"
	"ec graph.false rgb:431\n"
	"ec graph.trufae rgb:fd9\n"
	"ec graph.current rgb:0000ff\n"
	"ec graph.traced rgb:0000ff\n"
	"ec graph.box rgb:431\n"
	"ec graph.box2 rgb:ca6\n"
	"ec graph.box3 rgb:ca6\n"
	"ec graph.box4 rgb:ca6\n"
	"ec gui.cflow rgb:ffff00\n"
	"ec gui.dataoffset rgb:ffff00\n"
	"ec gui.background rgb:000000\n"
	"ec gui.background2 rgb:ffffff\n"
	"ec gui.border rgb:000000\n"
	"\n"
	"ec var rgb:fd9\n"
	"ec var.type rgb:850\n"
	"ec var.addr rgb:960\n"
	"ec var.name rgb:fd9\n"
	"\n"
	"ec ai.read rgb:cb7\n"
	"ec ai.write rgb:ca6\n"
	"ec ai.exec rgb:db7\n"
	"ec ai.seq rgb:960\n"
	"ec ai.ascii rgb:fec\n"
	"\n"
	"ec widget.bg rgb:db7 rgb:000000\n"
	"ec widget.sel rgb:000000 rgb:fec\n"
	"\n"
	"# Linux default colors (need to be sepiaized)\n"
	"\n"
	"ec diff.unknown rgb:881798\n"
	"ec diff.new rgb:c50f1f\n"
	"ec diff.match rgb:767676\n"
	"ec diff.unmatch rgb:c19c00\n"
;
static const char * const theme_smyck = \
	"# ref: http://color.smyck.org/\n"
	"ecd\n"
	"ec b0x00 rgb:222\n"
	"ec b0x7f rgb:fff\n"
	"ec b0xff rgb:eee\n"
	"ec other rgb:888\n"
	"ec bin rgb:888\n"
	"ec btext white\n"
	"ec ucall rgb:9f4\n"
	"ec call rgb:9f4\n"
	"ec ujmp rgb:9c4\n"
	"ec cjmp rgb:9c4\n"
	"ec jmp rgb:9c4\n"
	"ec cmp rgb:888\n"
	"ec comment rgb:9df\n"
	"ec usrcmt rgb:9f4\n"
	"ec ret rgb:d66\n"
	"ec num rgb:bb5\n"
	"ec fline rgb:2bc\n"
	"ec flow rgb:888\n"
	"ec flow2 rgb:ddd\n"
	"ec fname rgb:9df\n"
	"ec label rgb:9df\n"
	"ec math rgb:ddd\n"
	"ec mov rgb:ddd\n"
	"ec nop rgb:444\n"
	"ec offset rgb:bb5\n"
	"ec prompt rgb:bb5\n"
	"ec pop rgb:fd6\n"
	"ec push rgb:fd6\n"
	"ec reg rgb:fff\n"
	"ec args rgb:2bc\n"
	"ec help rgb:fd6\n"
	"ec input rgb:888\n"
	"ec trap rgb:d66\n"
	"ec swi rgb:d66\n"
	"ec creg rgb:df7\n"
	"ec flag rgb:9df\n"
	"ec linehl rgb:004\n"
	"ec ai.read rgb:66d\n"
	"ec ai.write rgb:d66\n"
	"ec ai.exec rgb:6d6\n"
	"ec ai.seq rgb:fd6\n"
	"ec ai.ascii white\n"
	"ec graph.box rgb:9df\n"
	"ec graph.box2 rgb:9df\n"
	"ec graph.box3 rgb:9df\n"
	"ec graph.box4 rgb:9df\n"
	"ec graph.true rgb:9c4\n"
	"ec graph.false rgb:d66\n"
	"ec graph.trufae rgb:9c4\n"
	"ec graph.current rgb:fff\n"
	"\n"
	"ec var rgb:9df\n"
	"ec var.type rgb:fd6\n"
	"ec var.addr rgb:bb5\n"
	"ec var.name rgb:9df\n"
	"\n"
	"ec widget.bg rgb:bb5 rgb:222\n"
	"ec widget.sel black rgb:9c4\n"
;
static const char * const theme_solarized = \
	"ecd\n"
	"ec b0x00 rgb:033\n"
	"ec b0x7f rgb:033\n"
	"ec b0xff rgb:577\n"
	"ec bin rgb:688\n"
	"ec btext rgb:899\n"
	"ec ucall rgb:890\n"
	"ec call rgb:890 #b80\n"
	"ec ujmp rgb:890\n"
	"ec cjmp rgb:890 # b80 #ee8\n"
	"ec jmp rgb:e37\n"
	"ec cmp rgb:c41 # ffe\n"
	"ec comment rgb:28d\n"
	"ec usrcmt rgb:890\n"
	"ec ret rgb:899\n"
	"ec fline rgb:b80\n"
	"ec flow rgb:3a9\n"
	"ec flow2 rgb:577\n"
	"ec fname rgb:d33\n"
	"ec label rgb:77c\n"
	"ec math rgb:28d\n"
	"ec mov rgb:3a9\n"
	"ec nop rgb:577\n"
	"ec num rgb:eee\n"
	"ec offset rgb:9aa\n"
	"ec other rgb:577\n"
	"ec pop rgb:688\n"
	"ec prompt rgb:c41\n"
	"ec push rgb:e37\n"
	"ec reg rgb:de8\n"
	"ec args rgb:de8\n"
	"ec help rgb:e37\n"
	"ec input rgb:c41\n"
	"ec trap rgb:f00\n"
	"ec swi rgb:f00\n"
	"ec creg rgb:de8\n"
	"ec flag rgb:c41\n"
	"ec linehl rgb:004\n"
	"\n"
	"ec graph.box rgb:033\n"
	"ec graph.box2 rgb:de8\n"
	"ec graph.box3 rgb:033\n"
	"ec graph.box4 rgb:033\n"
	"ec graph.true rgb:890\n"
	"ec graph.false rgb:c41\n"
	"ec graph.trufae rgb:899\n"
	"ec graph.current rgb:f00\n"
	"ec graph.traced red\n"
	"\n"
	"ec var rgb:28d\n"
	"ec var.type rgb:cc8\n"
	"ec var.addr rgb:9aa\n"
	"ec var.name rgb:28d\n"
	"\n"
	"ec ai.read rgb:39e\n"
	"ec ai.write rgb:577\n"
	"ec ai.exec rgb:ab4\n"
	"ec ai.seq rgb:899\n"
	"ec ai.ascii rgb:899\n"
	"\n"
	"# background of the terminal must be 134 o 033\n"
	"# xterm -fn 10x20 -bg rgb:00/30/30 -fg rgb:e0/e0/e0 -e \"r2 -c 'eco solarized' /bin/ls\"\n"
	"\n"
	"ec widget.bg rgb:bb5 rgb:033\n"
	"ec widget.sel white rgb:c41\n"
;
static const char * const theme_tango = \
	"# background of the terminal must be 134 o 033\n"
	"# xterm -fn 10x20 -bg rgb:20/20/20 -fg rgb:d0/d0/d0\n"
	"# ref: https://github.com/eugeneching/ida-consonance/blob/master/ida-consonance.clr\n"
	"\n"
	"#              LIGHT   MEDIUM  DARK\n"
	"# Butter 	fe5 	ed0 	ca0\n"
	"# Orange 	fa3 	f70 	c50\n"
	"# Chocolate 	eb6 	c71 	950\n"
	"# Chameleon 	8e3 	7d1 	590\n"
	"# Sky Blue 	79c 	36a 	248\n"
	"# Plum 		a7a 	757 	536\n"
	"# Scarlet Red 	e22 	c00 	a00\n"
	"# Aluminium 	eee 	ddd 	bbb\n"
	"# DarkAluminium	888 	555 	333\n"
	"\n"
	"# DARK THEME\n"
	"ecd\n"
	"ec b0x00 rgb:757\n"
	"ec b0x7f rgb:c50\n"
	"ec b0xff rgb:950\n"
	"ec btext rgb:ca0\n"
	"\n"
	"ec bin rgb:bbb\n"
	"ec ucall rgb:7d1\n"
	"ec call rgb:7d1\n"
	"ec ujmp rgb:370\n"
	"ec cjmp rgb:370\n"
	"ec jmp rgb:370\n"
	"ec cmp rgb:bbb\n"
	"ec comment rgb:370\n"
	"ec usrcmt rgb:7d1\n"
	"ec ret rgb:248\n"
	"ec fline rgb:950\n"
	"ec flow rgb:248\n"
	"ec flow2 rgb:248\n"
	"ec fname rgb:ca0\n"
	"ec label rgb:c50\n"
	"ec math rgb:bbb\n"
	"ec mov rgb:bbb\n"
	"ec nop rgb:333\n"
	"ec num rgb:ca0\n"
	"ec offset rgb:bbb\n"
	"ec other rgb:950\n"
	"ec pop rgb:757\n"
	"ec prompt rgb:950\n"
	"ec push rgb:757\n"
	"ec reg rgb:c50\n"
	"ec args rgb:ca0\n"
	"ec help rgb:ddd\n"
	"ec input rgb:ca0\n"
	"ec trap rgb:a00\n"
	"ec swi rgb:950\n"
	"ec creg rgb:590\n"
	"ec flag rgb:950\n"
	"ec linehl rgb:004\n"
	"\n"
	"ec graph.true rgb:370\n"
	"ec graph.false rgb:a00\n"
	"ec graph.trufae rgb:bbb\n"
	"ec graph.current rgb:7d1\n"
	"ec graph.traced rgb:f00\n"
	"ec graph.box rgb:950\n"
	"ec graph.box2 rgb:7d1\n"
	"ec graph.box3 rgb:950\n"
	"ec graph.box4 rgb:950\n"
	"\n"
	"ec var rgb:370\n"
	"ec var.type rgb:a41\n"
	"ec var.addr rgb:aaa\n"
	"ec var.name rgb:370\n"
	"\n"
	"ec ai.read rgb:a81\n"
	"ec ai.write rgb:a41\n"
	"ec ai.exec rgb:c50\n"
	"ec ai.seq rgb:ca0\n"
	"ec ai.ascii rgb:ca0\n"
	"\n"
	"ec widget.bg rgb:bb5 rgb:030\n"
	"ec widget.sel black rgb:840\n"
;
static const char * const theme_twilight = \
	"# background of the terminal must be 134 o 033\n"
	"# xterm -fn 10x20 -bg rgb:14/14/14 -fg rgb:fe/ff/d3\n"
	"# ref: https://github.com/eugeneching/ida-consonance/blob/master/ida-consonance.clr\n"
	"ecd\n"
	"ec b0x00 rgb:446\n"
	"ec b0x7f rgb:c64\n"
	"ec b0xff rgb:c64\n"
	"ec bin rgb:ffd\n"
	"ec other rgb:aaa\n"
	"ec btext rgb:ab7\n"
	"ec ucall rgb:ffd\n"
	"ec call rgb:ffd\n"
	"# rgb:788 is grayish better use for nops than calls\n"
	"# rgb:ab7 is greenish was for jumps but better highlight them with whitish (ffd)\n"
	"ec ujmp rgb:ffd\n"
	"ec cjmp rgb:ffd\n"
	"ec jmp rgb:ffd\n"
	"ec cmp rgb:ab7\n"
	"ec comment rgb:c64\n"
	"ec usrcmt rgb:9b7\n"
	"ec ret rgb:ab7\n"
	"ec fline rgb:ca6\n"
	"ec flow rgb:446\n"
	"ec flow2 rgb:446\n"
	"ec fname rgb:c64\n"
	"ec label rgb:788\n"
	"ec math rgb:aaa\n"
	"ec invalid rgb:c64\n"
	"ec mov rgb:aaa\n"
	"ec nop rgb:445\n"
	"ec num rgb:aaa\n"
	"ec offset rgb:788\n"
	"ec pop rgb:bb7\n"
	"ec prompt rgb:788\n"
	"ec push rgb:ffd\n"
	"ec reg rgb:ca6\n"
	"ec args rgb:bb7\n"
	"ec help rgb:ab7\n"
	"ec input rgb:ca6\n"
	"ec trap rgb:445\n"
	"ec swi rgb:bb7\n"
	"ec creg rgb:788\n"
	"ec flag rgb:ffd\n"
	"ec linehl rgb:004\n"
	"ec graph.true rgb:9b7\n"
	"ec graph.false rgb:b97\n"
	"ec graph.trufae rgb:64c\n"
	"ec graph.current blue\n"
	"ec graph.traced blue\n"
	"ec graph.box rgb:9b7\n"
	"ec graph.box2 rgb:b97\n"
	"ec graph.box3 rgb:b97\n"
	"ec graph.box4 rgb:b97\n"
	"\n"
	"ec var rgb:c64\n"
	"ec var.type rgb:aa6\n"
	"ec var.addr rgb:788\n"
	"ec var.name rgb:c64\n"
	"\n"
	"ec ai.read rgb:db7\n"
	"ec ai.write rgb:ca6\n"
	"ec ai.exec rgb:ffd\n"
	"ec ai.seq rgb:bb7\n"
	"ec ai.ascii rgb:bb7\n"
	"\n"
	"ec widget.bg rgb:ffd rgb:344\n"
	"ec widget.sel black rgb:ca6\n"
;
static const char * const theme_white = \
	"ecd\n"
	"ec label rgb:c60\n"
	"ec help blue\n"
	"ec math blue\n"
	"ec bin blue\n"
	"ec crypto magenta\n"
	"ec other blue\n"
	"ec push black\n"
	"ec pop black\n"
	"ec cmp magenta\n"
	"ec nop black\n"
	"ec jmp blue\n"
	"ec cjmp magenta\n"
	"ec ujmp green\n"
	"ec call magenta\n"
	"ec ucall magenta\n"
	"ec input black\n"
	"ec fline green\n"
	"ec prompt black\n"
	"ec flag blue\n"
	"ec fname blue\n"
	"ec comment magenta\n"
	"ec usrcmt green\n"
	"ec reg blue\n"
	"ec creg blue\n"
	"ec trap red\n"
	"ec swi green\n"
	"ec num blue\n"
	"ec b0x7f gray\n"
	"ec b0xff red\n"
	"ec b0x00 rgb:777\n"
	"ec btext black\n"
	"ec offset black\n"
	"ec flow blue\n"
	"ec flow2 green\n"
	"ec mov black\n"
	"ec invalid red\n"
	"ec linehl rgb:ff0\n"
	"ec graph.box blue\n"
	"ec graph.box2 green\n"
	"ec graph.box3 red\n"
	"ec graph.box4 blue\n"
	"ec graph.true green\n"
	"ec graph.false red\n"
	"ec graph.trufae blue\n"
	"ec graph.current rgb:0ff\n"
	"ec graph.traced red\n"
	"\n"
	"ec var magenta\n"
	"ec var.type blue\n"
	"ec var.addr magenta\n"
	"ec var.name magenta\n"
	"\n"
	"ec ai.read magenta\n"
	"ec ai.write magenta\n"
	"ec ai.exec blue\n"
	"ec ai.seq black\n"
	"ec ai.ascii black\n"
	"\n"
	"ec widget.bg blue white\n"
	"ec widget.sel black yellow\n"
;
static const char * const theme_white2 = \
	"ecd\n"
	"ec ai.exec rgb:6d6\n"
	"ec ai.read rgb:66d\n"
	"ec ai.write rgb:d66\n"
	"ec args rgb:2bc\n"
	"ec b0x00 rgb:878585             # 00 bytes\n"
	"ec b0x7f rgb:fff\n"
	"ec b0xff rgb:abb0b6             # ff bytes\n"
	"ec bin rgb:4F1900 0 bold               # operations like xor\n"
	"ec btext rgb:66737f             # text section opcode? first byte of opcode\n"
	"ec call rgb:A63A00 0 bold              # call instructions\n"
	"ec cjmp rgb:376B4C 0 bold underline              # je, jg, etc\n"
	"ec cmp rgb:F92672 0 bold               # cmp instructions\n"
	"ec comment rgb:001296 0 italic           # r2 renerated comments\n"
	"ec creg rgb:025F49 rgb:BFFFCB              # modified registries\n"
	"ec flag rgb:912D3A 0 bold              # flags. usually apprears above the current seek\n"
	"ec fline rgb:8D0004             # function line. the boundary of a function\n"
	"ec flow rgb:376B4C              # lines that show jump destinations\n"
	"ec flow2 rgb:ddd\n"
	"ec fname rgb:8D0004 0 bold italic             # function name\n"
	"ec graph.box rgb:007D96         # unselected boxes in visual graph mode\n"
	"ec graph.box2 rgb:D100A7        # currently selected box in visual graph mode\n"
	"ec graph.box3 rgb:66d9ef\n"
	"ec graph.box4 rgb:66d9ef\n"
	"ec graph.current rgb:fff\n"
	"ec graph.false rgb:990000       # false line on visual graph mode\n"
	"ec graph.true rgb:376B4C        # true line on visual graph mode\n"
	"ec graph.trufae rgb:007D96      # lines in graph that does not have true or false\n"
	"ec help rgb:2F6F9F 0 italic              # color of help explanation text\n"
	"ec input rgb:888\n"
	"ec jmp rgb:376B4C 0 bold underline               # jmp instructions\n"
	"ec label rgb:66d9ef\n"
	"ec linehl rgb:42505e            # highlight color when using asm.highlight\n"
	"ec math rgb:761B48 0 bold              # math instructions like sub, add\n"
	"ec mov black\n"
	"ec nop rgb:a37acc               # nop instructions\n"
	"ec num rgb:761B48               # numbers\n"
	"ec offset rgb:2F6F9F            # offset in visual mode\n"
	"ec other rgb:4D4D4D\n"
	"ec pop rgb:ff7733               # pop instructions\n"
	"ec prompt rgb:2F6F9F 0 bold            # color for main r2 prompt\n"
	"ec push rgb:ff7733              # push instructions\n"
	"ec reg rgb:400080               # register names\n"
	"ec ret rgb:F92672 0 bold                  # ret opcode\n"
	"ec swi rgb:d66\n"
	"ec trap rgb:d66\n"
	"ec ucall rgb:A63A00 0 bold\n"
	"ec ujmp rgb:376B4C 0 bold underline\n"
	"ec usrcmt rgb:007D96 0 bold           # user comment\n"
	"\n"
	"ec var rgb:400080      # var name in function\n"
	"ec var.type rgb:305F65 0 italic     # function variable\n"
	"ec var.addr magenta\n"
	"ec var.name rgb:400080\n"
	"\n"
	"ec widget.bg rgb:2f6f9f white\n"
	"ec widget.sel white rgb:f73 bold\n"
;
static const char * const theme_zenburn = \
	"# background of the terminal must be 134 o 033 # xterm -fn 10x20 -bg rgb:20/20/20 -fg rgb:ff/ff/c7\n"
	"# ref: https://github.com/eugeneching/ida-consonance/blob/master/ida-consonance.clr\n"
	"ecd\n"
	"ec b0x00 rgb:8a8\n"
	"ec b0x7f rgb:aaa\n"
	"ec b0xff rgb:8a8\n"
	"ec bin rgb:fcb\n"
	"ec other rgb:aaa\n"
	"ec btext rgb:ddc\n"
	"ec ucall rgb:d52\n"
	"ec call rgb:d52\n"
	"ec ujmp rgb:d52\n"
	"ec cjmp rgb:d52\n"
	"ec jmp rgb:d52\n"
	"ec cmp rgb:ff8\n"
	"ec comment rgb:7a7\n"
	"ec usrcmt rgb:d52\n"
	"ec ret rgb:ff8\n"
	"ec fline rgb:ddc\n"
	"ec flow rgb:d50\n"
	"ec flow2 rgb:d50\n"
	"ec fname rgb:d50\n"
	"ec label rgb:d50\n"
	"ec math rgb:daa\n"
	"ec invalid rgb:ff8\n"
	"ec mov rgb:cd9\n"
	"ec nop rgb:8a8\n"
	"ec num rgb:c88\n"
	"ec offset rgb:aaa\n"
	"ec pop rgb:ffc\n"
	"ec prompt rgb:d50\n"
	"ec push rgb:ffc\n"
	"ec reg rgb:8dd\n"
	"ec args rgb:7a7\n"
	"ec help rgb:ffc\n"
	"ec input rgb:da8\n"
	"ec trap rgb:d22\n"
	"ec swi rgb:d22\n"
	"ec creg rgb:ccc\n"
	"ec flag rgb:d50\n"
	"ec linehl rgb:004\n"
	"\n"
	"ec graph.box rgb:aaa\n"
	"ec graph.box2 rgb:d52\n"
	"ec graph.box3 rgb:aaa\n"
	"ec graph.box4 rgb:aaa\n"
	"ec graph.true rgb:aaa\n"
	"ec graph.false rgb:d52\n"
	"ec graph.trufae rgb:aaa\n"
	"ec graph.current blue\n"
	"ec graph.traced red\n"
	"\n"
	"ec var rgb:7a7\n"
	"ec var.type rgb:eec\n"
	"ec var.addr rgb:aaa\n"
	"ec var.name rgb:7a7\n"
	"\n"
	"ec ai.read rgb:f87\n"
	"ec ai.write rgb:c88\n"
	"ec ai.exec rgb:cd9\n"
	"ec ai.seq rgb:ddc\n"
	"ec ai.ascii rgb:ddc\n"
	"\n"
	"ec widget.bg rgb:8dd rgb:333\n"
	"ec widget.sel white rgb:b52 bold\n"
;
RConsTheme d_themes[] = {
	{ "ayu", theme_ayu },
	{ "basic", theme_basic },
	{ "behelit", theme_behelit },
	{ "blood", theme_blood },
	{ "bluy", theme_bluy },
	{ "bobcrawl", theme_bobcrawl },
	{ "bold", theme_bold },
	{ "bright", theme_bright },
	{ "cga", theme_cga },
	{ "chrome", theme_chrome },
	{ "consonance", theme_consonance },
	{ "dark", theme_dark },
	{ "darkda", theme_darkda },
	{ "focus", theme_focus },
	{ "gb", theme_gb },
	{ "gentoo", theme_gentoo },
	{ "greepy", theme_greepy },
	{ "gruvbox", theme_gruvbox },
	{ "iaito", theme_iaito },
	{ "lima", theme_lima },
	{ "matrix", theme_matrix },
	{ "monokai", theme_monokai },
	{ "ogray", theme_ogray },
	{ "onedark", theme_onedark },
	{ "pink", theme_pink },
	{ "rasta", theme_rasta },
	{ "sepia", theme_sepia },
	{ "smyck", theme_smyck },
	{ "solarized", theme_solarized },
	{ "tango", theme_tango },
	{ "twilight", theme_twilight },
	{ "white", theme_white },
	{ "white2", theme_white2 },
	{ "zenburn", theme_zenburn },
	{0,0}
};
