/* c55plus - LGPL - Copyright 2013-2015 - th0rpe */
/* this .c is included , not compiled */
#include "hashvector.h"

extern st32 get_hashfunc_01(st32 A1, st32 A2);
extern st32 get_hashfunc_02(st32 A1, st32 A2);
extern st32 get_hashfunc_03(st32 A1, st32 A2);
extern st32 get_hashfunc_04(st32 A1, st32 A2);
extern st32 get_hashfunc_05(st32 A1, st32 A2);
extern st32 get_hashfunc_06(st32 A1, st32 A2);
extern st32 get_hashfunc_07(st32 A1, st32 A2);
extern st32 get_hashfunc_08(st32 A1, st32 A2);
extern st32 get_hashfunc_09(st32 A1, st32 A2);
extern st32 get_hashfunc_10(st32 A1, st32 A2);
extern st32 get_hashfunc_11(st32 A1, st32 A2);
extern st32 get_hashfunc_12(st32 A1, st32 A2);
extern st32 get_hashfunc_13(st32 A1, st32 A2);
extern st32 get_hashfunc_14(st32 A1, st32 A2);
extern st32 get_hashfunc_15(st32 A1, st32 A2);
extern st32 get_hashfunc_16(st32 A1, st32 A2);
extern st32 get_hashfunc_17(st32 A1, st32 A2);
extern st32 get_hashfunc_18(st32 A1, st32 A2);
extern st32 get_hashfunc_19(st32 A1, st32 A2);
extern st32 get_hashfunc_20(st32 A1, st32 A2);
extern st32 get_hashfunc_21(st32 A1, st32 A2);
extern st32 get_hashfunc_22(st32 A1, st32 A2);
extern st32 get_hashfunc_23(st32 A1, st32 A2);
extern st32 get_hashfunc_24(st32 A1, st32 A2);
extern st32 get_hashfunc_25(st32 A1, st32 A2);
extern st32 get_hashfunc_26(st32 A1, st32 A2);
extern st32 get_hashfunc_27(st32 A1, st32 A2);
extern st32 get_hashfunc_28(st32 A1, st32 A2);
extern st32 get_hashfunc_29(st32 A1, st32 A2);
extern st32 get_hashfunc_30(st32 A1, st32 A2);
extern st32 get_hashfunc_31(st32 A1, st32 A2);
extern st32 get_hashfunc_32(st32 A1, st32 A2);
extern st32 get_hashfunc_33(st32 A1, st32 A2);
extern st32 get_hashfunc_34(st32 A1, st32 A2);
extern st32 get_hashfunc_35(st32 A1, st32 A2);
extern st32 get_hashfunc_36(st32 A1, st32 A2);
extern st32 get_hashfunc_37(st32 A1, st32 A2);
extern st32 get_hashfunc_38(st32 A1, st32 A2);
extern st32 get_hashfunc_39(st32 A1, st32 A2);
extern st32 get_hashfunc_40(st32 A1, st32 A2);
extern st32 get_hashfunc_41(st32 A1, st32 A2);
extern st32 get_hashfunc_42(st32 A1, st32 A2);
extern st32 get_hashfunc_43(st32 A1, st32 A2);
extern st32 get_hashfunc_44(st32 A1, st32 A2);
extern st32 get_hashfunc_45(st32 A1, st32 A2);
extern st32 get_hashfunc_46(st32 A1, st32 A2);
extern st32 get_hashfunc_47(st32 A1, st32 A2);
extern st32 get_hashfunc_48(st32 A1, st32 A2);
extern st32 get_hashfunc_49(st32 A1, st32 A2);
extern st32 get_hashfunc_50(st32 A1, st32 A2);
extern st32 get_hashfunc_51(st32 A1, st32 A2);
extern st32 get_hashfunc_52(st32 A1, st32 A2);
extern st32 get_hashfunc_53(st32 A1, st32 A2);
extern st32 get_hashfunc_54(st32 A1, st32 A2);
extern st32 get_hashfunc_55(st32 A1, st32 A2);
extern st32 get_hashfunc_56(st32 A1, st32 A2);
extern st32 get_hashfunc_57(st32 A1, st32 A2);
extern st32 get_hashfunc_58(st32 A1, st32 A2);
extern st32 get_hashfunc_59(st32 A1, st32 A2);
extern st32 get_hashfunc_60(st32 A1, st32 A2);
extern st32 get_hashfunc_61(st32 A1, st32 A2);
extern st32 get_hashfunc_62(st32 A1, st32 A2);
extern st32 get_hashfunc_63(st32 A1, st32 A2);
extern st32 get_hashfunc_64(st32 A1, st32 A2);
extern st32 get_hashfunc_65(st32 A1, st32 A2);
extern st32 get_hashfunc_66(st32 A1, st32 A2);
extern st32 get_hashfunc_67(st32 A1, st32 A2);
extern st32 get_hashfunc_68(st32 A1, st32 A2);
extern st32 get_hashfunc_69(st32 A1, st32 A2);
extern st32 get_hashfunc_70(st32 A1, st32 A2);
extern st32 get_hashfunc_71(st32 A1, st32 A2);
extern st32 get_hashfunc_72(st32 A1, st32 A2);
extern st32 get_hashfunc_73(st32 A1, st32 A2);
extern st32 get_hashfunc_74(st32 A1, st32 A2);
extern st32 get_hashfunc_75(st32 A1, st32 A2);
extern st32 get_hashfunc_76(st32 A1, st32 A2);
extern st32 get_hashfunc_77(st32 A1, st32 A2);
extern st32 get_hashfunc_78(st32 A1, st32 A2);
extern st32 get_hashfunc_79(st32 A1, st32 A2);
extern st32 get_hashfunc_80(st32 A1, st32 A2);
extern st32 get_hashfunc_81(st32 A1, st32 A2);
extern st32 get_hashfunc_82(st32 A1, st32 A2);
extern st32 get_hashfunc_83(st32 A1, st32 A2);
extern st32 get_hashfunc_84(st32 A1, st32 A2);
extern st32 get_hashfunc_85(st32 A1, st32 A2);
extern st32 get_hashfunc_86(st32 A1, st32 A2);
extern st32 get_hashfunc_87(st32 A1, st32 A2);
extern st32 get_hashfunc_88(st32 A1, st32 A2);
extern st32 get_hashfunc_89(st32 A1, st32 A2);
extern st32 get_hashfunc_90(st32 A1, st32 A2);
extern st32 get_hashfunc_91(st32 A1, st32 A2);
