/* Automatically generated by configure - do not modify */
#ifndef CONFIG_TCCDIR
# define CONFIG_TCCDIR "/usr/local/lib/tcc"
#endif
#define GCC_MAJOR 4
#define GCC_MINOR 2
#define HOST_I386 1
#define TCC_VERSION "0.9.26"
