void kvx_decode_none(operand_t *, const ut32 *);
void kv3_decode_rw(operand_t *, const ut32 *);
void kv3_decode_ry(operand_t *, const ut32 *);
void kv3_decode_rz(operand_t *, const ut32 *);
void kv3_decode_ra(operand_t *, const ut32 *);
void kv3_decode_rap(operand_t *, const ut32 *);
void kv3_decode_raq(operand_t *, const ut32 *);
void kv3_decode_rb(operand_t *, const ut32 *);
void kv3_decode_rb_odd(operand_t *, const ut32 *);
void kv3_decode_rb_even(operand_t *, const ut32 *);
void kv3_decode_rbp(operand_t *, const ut32 *);
void kv3_decode_rbq(operand_t *, const ut32 *);
void kv3_decode_rc(operand_t *, const ut32 *);
void kv3_decode_rc_odd(operand_t *, const ut32 *);
void kv3_decode_rc_even(operand_t *, const ut32 *);
void kv3_decode_rd(operand_t *, const ut32 *);
void kv3_decode_rn(operand_t *, const ut32 *);
void kv3_decode_rm(operand_t *, const ut32 *);
void kv3_decode_u6(operand_t *, const ut32 *);
void kv3_decode_s10(operand_t *, const ut32 *);
void kv3_decode_s16(operand_t *, const ut32 *);
void kv3_decode_sys(operand_t *, const ut32 *);
void kv3_decode_imm32(operand_t *, const ut32 *);
void kv3_decode_imm37(operand_t *, const ut32 *);
void kv3_decode_imm43(operand_t *, const ut32 *);
void kv3_decode_imm64(operand_t *, const ut32 *);
void kv3_decode_off27(operand_t *, const ut32 *);
void kv3_decode_off54(operand_t *, const ut32 *);
void kv3_decode_pcrel17(operand_t *, const ut32 *);
void kv3_decode_pcrel27(operand_t *, const ut32 *);
void kv3_decode_shift(operand_t *, const ut32 *);
void kv3_decode_start_bit(operand_t *, const ut32 *);
void kv3_decode_stop_bit(operand_t *, const ut32 *);
void kv3_decode_ru(operand_t *, const ut32 *);
void kv3_decode_rt(operand_t *, const ut32 *);
void kv3_decode_rbq(operand_t *, const ut32 *);
void kv3_decode_rq(operand_t *, const ut32 *);
void kv3_decode_rp(operand_t *, const ut32 *);
void kv3_decode_ro(operand_t *, const ut32 *);
void kv3_decode_rs(operand_t *, const ut32 *);
void kv3_decode_rv(operand_t *, const ut32 *);
void kv3_decode_rr(operand_t *, const ut32 *);
void kv3_decode_re(operand_t *, const ut32 *);
