#include "opcode_10.c"
#include "opcode_11.c"
#include "opcode_12.c"
#include "opcode_13.c"
#include "opcode_14.c"
#include "opcode_15.c"
#include "opcode_16.c"
#include "opcode_20.c"
#include "opcode_21.c"
#include "opcode_22.c"
#include "opcode_23.c"
#include "opcode_24.c"
#include "opcode_25.c"
#include "opcode_26.c"
#include "opcode_27.c"
#include "opcode_2x.c"
#include "opcode_30.c"
#include "opcode_31.c"
#include "opcode_32.c"
#include "opcode_33.c"
#include "opcode_34.c"
#include "opcode_35.c"
#include "opcode_36.c"
#include "opcode_37.c"
#include "opcode_38.c"
#include "opcode_39.c"
#include "opcode_3x.c"
